import re
from typing import Any

from juriscraper.AbstractSite import logger
from juriscraper.lib.html_utils import fix_links_in_lxml_tree
from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    citation_regex = re.compile(
        r"Cite\s+as\s+(?P<citation>\d+ +Neb\.( App\.)? \d+)"
    )

    lower_court_regex = re.compile(
        r"""
                    Appeals?\s+from\s+the\s+
                    (?P<lower_court>.*?):\s+(?P<judge>.+?),\s*Judge
                    """,
        re.X | re.DOTALL,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = (
            "https://supremecourt.nebraska.gov/courts/supreme-court/opinions"
        )
        self.should_have_results = True

    def _return_response_text_object(self):
        """Remove faulty URLs from HTML

        Override _return_response_text_object in Abstract Site to remove any
        url to [site:url-brief] which causes lxml/ipaddress to explode.

        This bug only exists in 3.11 and 3.12

        <a href="https://[site:url-brief]/node/19063" ... >District Map</a>

        :return: The cleaned html tree of the site
        """
        if self.request["response"]:
            payload = self.request["response"].text
            text = self._clean_text(payload)
            html_tree = self._make_html_tree(text)

            for tag in html_tree.xpath(
                "//a[contains(@href, 'site:url-brief')]"
            ):
                parent = tag.getparent()
                if parent is not None:
                    parent.remove(tag)

            if hasattr(html_tree, "rewrite_links"):
                html_tree.rewrite_links(
                    fix_links_in_lxml_tree, base_href=self.request["url"]
                )
            return html_tree

    def _process_html(self):
        for table in self.html.xpath(".//table"):
            date_tags = table.xpath("preceding::time[1]/text()")
            if not date_tags:
                continue
            date = date_tags[0]

            for row in table.xpath(".//tr[td]"):
                c1, c2, c3 = row.xpath(".//td")
                docket = c1.xpath(".//text()")[0].strip()
                if "A-XX-XXXX" in docket or not c3.xpath(".//a"):
                    logger.info("Skip row %s", row.text_content())
                    continue

                citation = c2.xpath(".//text()")[0].strip()
                name = c3.xpath(".//a/text()")[0].strip()
                url = c3.xpath(".//a")[0].get("href")
                # This URL location is used for unpublished opinions
                if "/sites/default/files" in url:
                    status = "Unpublished"
                else:
                    status = "Published"

                self.cases.append(
                    {
                        "date": date,
                        "docket": docket,
                        "name": name,
                        "citation": citation,
                        "url": url,
                        "status": status,
                    }
                )

    def extract_from_text(self, scraped_text: str) -> dict[str, Any]:
        """Get the state citation and lower court info from the document's text."""
        result = {}

        match = self.citation_regex.search(scraped_text[:1000])
        if match:
            result["Citation"] = match.group("citation")

        match = self.lower_court_regex.search(scraped_text)
        if match:
            result["Docket"] = {
                "appeal_from_str": re.sub(
                    r"\s+", " ", match.group("lower_court")
                ).strip()
            }
            result["OriginatingCourtInformation"] = {
                "assigned_to_str": titlecase(
                    re.sub(r"\s+", " ", match.group("judge")).strip()
                )
            }

        return result



"""
Scraper for the Supreme Court of Tennessee
CourtID: tenn
Court Short Name: Tenn.
"""

import re
from datetime import date, datetime
from urllib.parse import urljoin

from juriscraper.lib.type_utils import OpinionType
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://www.tncourts.gov/courts/supreme-court/opinions"
        self.court_id = self.__module__
        self.status = "Unknown"
        self.should_have_results = True
        self.first_opinion_date = datetime(1993, 1, 22)
        self.days_interval = 7
        self.make_backscrape_iterable(kwargs)

    def _process_html(self):
        """
        Parse the HTML table rows and extract case details.

        Iterates over each table row in the HTML, extracting the date, URL, case name,
        docket number, judge, lower court judge, summary, per curiam status, and opinion type.
        Appends a dictionary with these details to self.cases.
        """
        for row in self.html.xpath("//tr"):
            date = (
                row.xpath(
                    ".//td[contains(@class, 'views-field-field-opinions-date-filed')]"
                )[0]
                .text_content()
                .strip()
            )
            section = row.xpath(
                ".//td[contains(@class, 'views-field-field-opinions-case-number')]"
            )[0]
            url = section.xpath(".//a")[0].get("href")

            name_text = section.xpath(".//a")[0].text_content()
            type_match = re.search(r"\(([^)]+)\)", name_text)
            type_raw = type_match.group(1).lower() if type_match else ""

            opinion_type = self.extract_type(type_raw)

            name = re.sub(r"\s*\([^)]+\)", "", name_text).strip()
            rows = [
                row.strip()
                for row in section.text_content().strip().split("\n", 4)
            ]

            judge = (
                rows[2].split(": ")[1] if "Authoring Judge" in rows[2] else ""
            )
            lower_court_judge = (
                rows[3].split(": ")[1]
                if "Trial Court Judge" in rows[3]
                else ""
            )
            summary = rows[-1] if "Judge" not in rows[-1] else ""
            per_curiam = False
            if "curiam" in judge.lower():
                judge = ""
                per_curiam = True

            self.cases.append(
                {
                    "date": date,
                    "url": url,
                    "name": name,
                    "docket": rows[1],
                    "judge": judge,
                    "lower_court_judge": lower_court_judge,
                    "summary": summary,
                    "per_curiam": per_curiam,
                    "type": opinion_type,
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract precedential_status and appeal_from_str from scraped text.

        :param scraped_text: Text of the scraped content
        :return: dictionary with precedential_status and appeal_from_str
        """
        lower_court = self.extract_lower_court_name(scraped_text)
        precedential_status = (
            "Published"
            if "MEMORANDUM OPINION" not in scraped_text
            else "Unpublished"
        )
        result = {
            "OpinionCluster": {"precedential_status": precedential_status}
        }
        if lower_court:
            result["Docket"] = {"appeal_from_str": lower_court}
        return result

    def extract_lower_court_name(self, text: str) -> str:
        """Extract the lower court name from the provided opinion text.

        Sometimes there is no introductory cue for the lower court. In that
        case, it's usually between the case name that contains a "v." and the
        docket number "No."
        That's what the third pattern is for

        :param text: the opinion's extracted text
        :return: the lower court or an empty string
        """

        patterns = [
            r"Appeal by Permission from.+\n(.+)\n",
            r"Appeal from the (.+)",
            r".+\bv\..+\n(?P<lower_court>.*(\n.*)?)\n.+No\.",
        ]

        for pattern in patterns:
            match = re.search(pattern, text[:1000])
            if match:
                return match.group(1).strip()

        return ""

    def extract_type(self, type_raw: str) -> str:
        """
        Map a raw opinion type string to a standardized type.

        :param type_raw: Raw type string extracted from the opinion (e.g., 'concurring', 'dissenting')
        :return: Standardized type string from types_mapping
        """
        if "concurring" in type_raw:
            op_type = OpinionType.CONCURRENCE
        elif "in part" in type_raw:
            op_type = OpinionType.CONCURRING_IN_PART_AND_DISSENTING_IN_PART
        elif "dissenting" in type_raw:
            op_type = OpinionType.DISSENT
        else:
            op_type = OpinionType.MAJORITY

        return op_type.value

    def _download_backwards(self, dates: tuple[date, date]) -> None:
        r"""Download cases within a given date range.

        :param dates: Tuple containing the start and end dates \(`date`, `date`\) for the query.
        :return None:
        """
        start, end = dates
        self.url = urljoin(
            self.url,
            f"?field_opinions_date_filed={start.strftime('%Y-%m-%d')}&field_opinions_date_filed_1={end.strftime('%Y-%m-%d')}",
        )
        self.html = self._download()
        self._process_html()



"""
History:
    - 2014-08-05: Updated by mlr because it was not working, however, in middle
      of update, site appeared to change. At first there were about five
      columns in the table and scraper was failing. Soon, there were seven and
      the scraper started working without my fixing it. Very odd.
    - 2023-01-13: Update to use RSS Feed
    - 2025-08-27: Added extract_from_text and avoid duplicates.
"""

import re

import feedparser
from lxml.html import tostring

from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.ca9.uscourts.gov/opinions/index.xml"
        self.status = "Published"
        self.should_have_results = True

    def _process_html(self) -> None:
        feed = feedparser.parse(tostring(self.html))
        for item in feed["entries"]:
            case = {
                "name": titlecase(item["title"]),
                "url": item["link"],
                "date": item["published"],
                "status": self.status,
                "docket": item["link"].split("/")[-1][:-4],
            }
            # Avoid duplicates cases from source page
            if case not in self.cases:
                self.cases.append(case)

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """
        pattern = re.compile(
            r"""
            (?:
               Appeals?\s+from\s+the\s+
              | On\s+Remand\s+from\s+the\s+
              | On\s+Petition\s+for\s+Review\s+of\s+an\s+Order\s+of\s+the\s+
            )
            (?P<lower_court>.+(?:\n.+)?)
            """,
            re.X,
        )

        lower_court = ""
        if match := pattern.search(scraped_text):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()

        if lower_court:
            return {
                "Docket": {
                    "appeal_from_str": lower_court,
                }
            }

        return {}



from datetime import date, datetime
from urllib.parse import urlencode

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    # This URL will show most recent opinions
    base_url = "https://www.ca1.uscourts.gov/opn/aci"
    days_interval = 5
    first_opinion_date = datetime(2003, 3, 23)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = self.base_url
        self.make_backscrape_iterable(kwargs)
        self.should_have_results = True

    def _process_html(self):
        for row in self.html.xpath("//tr[not(th)]"):
            title = row.xpath("td[2]/a/text()")[0]
            url = row.xpath("td[2]/a/@href")[0]
            status = self.get_status_from_opinion_title(title)
            docket = row.xpath("td[3]/a/text()")[0]
            date_filed = row.xpath("td[1]/span/text()")[0]
            name = row.xpath("td[4]/text()")[0]
            lower_court = row.xpath("td[4]/span/text()")[0]
            self.cases.append(
                {
                    "name": name.strip(),
                    "url": url,
                    "date": date_filed,
                    "status": status,
                    "docket": docket,
                    "lower_court": lower_court,
                }
            )

    def get_status_from_opinion_title(self, title: str) -> str:
        """Status is encoded in opinion's link title

        :param title: opinion title. Ex: 23-1667P.01A, 23-1639U.01A

        :return: status string
        """
        if "U" in title:
            status = "Unpublished"
        elif "P" in title:
            status = "Published"
        elif "E" in title:
            status = "Errata"
        else:
            status = "Unknown"
        return status

    def _download_backwards(self, dates: tuple[date]) -> None:
        """Change URL to backscraping date range

        :param dates: tuple with date range to scrape
        :return None
        """
        start, end = dates
        logger.info("Backscraping for range %s %s", *dates)
        params = {
            "field_opn_csno_value_op": "starts",
            "field_opn_issdate_value[min][date]": start.strftime("%m/%d/%Y"),
            "field_opn_issdate_value[max][date]": end.strftime("%m/%d/%Y"),
        }
        self.url = f"{self.base_url}?{urlencode(params)}"
        self.html = self._download()
        self._process_html()



"""Scraper for North Carolina Court of Appeals
CourtID: ncctapp
Court Short Name: N.C. Ct. App.
Author: Jon Andersen
History:
    2014-08-04: Created by Jon Andersen
"""

from juriscraper.opinions.united_states.state import nc


class Site(nc.Site):
    court = "coa"
    court_number = 2



"""Scraper for the Supreme Court of Delaware
CourtID: del

Creator: Andrei Chelaru
Reviewer: mlr
"""

import re

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = (
            "http://courts.delaware.gov/opinions/list.aspx?ag=supreme%20court"
        )
        # Note that we can't do the usual thing here because 'del' is a Python keyword.
        self.court_id = "juriscraper.opinions.united_states.state.del"
        self.should_have_results = True
        self.status = "Published"

    def _process_html(self):
        """Process the html and extract out the opinions

        :return: None
        """
        for row in self.html.xpath("//table/tr"):
            self.cases.append(
                {
                    "name": row.xpath("./td[1]/a/text()")[0].strip(),
                    "date": row.xpath("./td[2]/text()")[0].strip(),
                    "docket": row.xpath("./td[3]/a/text()")[0].strip(),
                    "url": row.xpath("./td[1]/a/@href")[0].strip(),
                    "judge": " ".join(row.xpath("./td[6]/text()")),
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """

        cleaned_text = "\n".join(
            line.rsplit("§", 1)[-1] if "§" in line else line
            for line in scraped_text[:1500].splitlines()
        )

        pattern = re.compile(
            r"Court\s+Below[–—\-:]\s*(?:the\s+)?(?P<lower_court>.*?)\n\s*\n",
            re.X | re.IGNORECASE | re.DOTALL,
        )

        number_pattern = re.compile(
            r"Nos?\.:?\s*(?P<lower_court_number>\S{5,})(?:\s|\()",
            re.IGNORECASE,
        )

        result = {}

        if match := pattern.search(cleaned_text):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()
            result["Docket"] = {"appeal_from_str": lower_court}

            if number_match := number_pattern.search(cleaned_text):
                lower_court_number = number_match.group(
                    "lower_court_number"
                ).strip()
                result["OriginatingCourtInformation"] = {
                    "docket_number": lower_court_number
                }

        return result



from .fisc import Site as FiscSite


class Site(FiscSite):
    def skip_record(self, docket: str, filename: str) -> bool:
        """Skips a record that belongs to FISC

        :param docket: docket number
        :param filename: opinion title

        :return: True if record should skipped
        """
        return "FISCR" not in docket and "FISCR" not in filename



import datetime
import hashlib
import hmac
import os

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSite import OpinionSite


def set_api_token_header(site: OpinionSite) -> None:
    """
    Puts the NY_API_TOKEN in the X-Api-Token header
    Creates the Site.headers attribute, copying the
    scraper_site.request[headers]

    :param scraper_site: a Site Object
    :returns: None
    """
    if site.test_mode_enabled():
        return
    api_token = os.environ.get("NY_API_TOKEN", None)
    if not api_token:
        logger.warning(
            "NY_API_TOKEN environment variable is not set. "
            f"It is required for scraping New York Court: {site.court_id}"
        )
        return
    site.request["headers"]["X-APIKEY"] = api_token
    site.needs_special_headers = True


def generate_aws_sigv4_headers(
    payload: str,
    table_name: str = None,
    creds: dict = None,
    signed_headers: str = "host;x-amz-date;x-amz-security-token;x-amz-target",
    target: str = "DynamoDB_20120810.Scan",
    service: str = "dynamodb",
    region: str = "us-west-2",
):
    """Generate AWS Signature Version 4 headers for a DynamoDB Scan request.

    This function builds the necessary SigV4 signing information and returns
    a complete set of HTTP headers you can pass to `requests` (or any HTTP
    client) to perform an authenticated DynamoDB Scan.
    :param payload: The JSON payload to send in the request body
    :param table_name: Table to scan or query
    :param creds: temp dictionary of credentials
    :param signed_headers: headers to include
    :param target: The request
    :param service: type of query
    :param region: the aws region
    :return: signed headers to use to query db
    """
    # 1. Prepare SigV4 signing
    now = datetime.datetime.utcnow()
    amz_date = now.strftime("%Y%m%dT%H%M%SZ")
    date_stamp = now.strftime("%Y%m%d")
    payload_hash = hashlib.sha256(payload.encode()).hexdigest()
    host = f"{service}.{region}.amazonaws.com"

    # Canonical headers
    canonical_headers = []
    headers_dict = {
        "content-type": "application/x-amz-json-1.0",
        "host": host,
        "x-amz-content-sha256": payload_hash,
        "x-amz-date": amz_date,
        "x-amz-security-token": creds["SessionToken"],
        "x-amz-target": target,
    }
    for h in sorted(signed_headers.split(";")):
        canonical_headers.append(f"{h}:{headers_dict[h]}\n")
    canonical_headers_str = "".join(canonical_headers)

    # Create canonical request
    canonical_request = "\n".join(
        ["POST", "/", "", canonical_headers_str, signed_headers, payload_hash]
    )

    algorithm = "AWS4-HMAC-SHA256"
    credential_scope = f"{date_stamp}/{region}/{service}/aws4_request"

    string_to_sign = "\n".join(
        [
            algorithm,
            amz_date,
            credential_scope,
            hashlib.sha256(canonical_request.encode()).hexdigest(),
        ]
    )

    # Signing key derivation
    def sign(key, msg):
        return hmac.new(key, msg.encode(), hashlib.sha256).digest()

    k_date = sign(("AWS4" + creds["SecretKey"]).encode(), date_stamp)
    k_region = sign(k_date, region)
    k_service = sign(k_region, service)
    k_signing = sign(k_service, "aws4_request")

    signature = hmac.new(
        k_signing, string_to_sign.encode(), hashlib.sha256
    ).hexdigest()

    return {
        "Content-Type": "application/x-amz-json-1.0",
        "X-Amz-Date": amz_date,
        "X-Amz-Security-Token": creds["SessionToken"],
        "X-Amz-Target": target,
        "Authorization": (
            f"{algorithm} Credential={creds['AccessKeyId']}/{credential_scope}, "
            f"SignedHeaders={signed_headers}, Signature={signature}"
        ),
        "X-Amz-Content-Sha256": payload_hash,
    }



from urllib.parse import urljoin

from juriscraper.AbstractSite import logger
from juriscraper.opinions.united_states.state import wis


class Site(wis.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.base_url = "https://www.wicourts.gov/other/appeals/caopin.jsp"
        self.set_url()
        self.cite_regex = r"20\d{2}\sWI App\s\d+"

    def combine_opinions(self, url: str, docket_number: str) -> bool:
        """Combine duplicate opinions in self.cases

        Wisconsin Court of Appeals generates a row for every docket number
        for combined cases.  If this is the case and the opinions all
        point to the same document, combine the docket numbers

        :return: True if opinion already in self.cases
        """
        for case in self.cases:
            if case["url"] == url:
                case["docket"] = f"{case['docket']}, {docket_number}"
                case["docket"] = ", ".join(sorted(case["docket"].split(", ")))
                return True
        return False

    def _process_html(self):
        """Process Wisconsin Ct of Appeals rows

        :return: None
        """
        for row in self.html.xpath(".//table/tbody/tr"):
            date, docket, caption, district, county, link = row.xpath("./td")
            long_caption = caption.text_content()
            if "[Decision/Opinion withdrawn" in long_caption:
                logger.debug("Skipping withdrawn opinion: %s", long_caption)
                continue

            captions = caption.xpath(".//text()")
            if "Errata" in captions[0]:
                status, name = "Errata", captions[1]
            elif "Recommended for Publication" in captions[-1]:
                status, name = "Published", captions[0]
            else:
                status, name = "Unpublished", captions[0]

            url = urljoin(
                "https://www.wicourts.gov", link.xpath("./input")[0].name
            )
            docket_number = docket.text
            if self.combine_opinions(url, docket_number):
                logger.debug("Duplicate row: %s", name)
                continue
            lower_court = f"Wisconsin Circuit Court, {county.text} County"

            self.cases.append(
                {
                    "date": date.text,
                    "name": name.strip(),
                    "url": url,
                    "docket": docket_number,
                    "status": status,
                    "lower_court": lower_court,
                }
            )



from enum import Enum


# OpinionType is used to represent the type of opinion in a case. These values are the ones used on CourtListener.
class OpinionType(Enum):
    UNANIMOUS = "015unamimous"
    MAJORITY = "020lead"
    PLURALITY = "025plurality"
    CONCURRENCE = "030concurrence"
    CONCURRING_IN_PART_AND_DISSENTING_IN_PART = "035concurrenceinpart"
    DISSENT = "040dissent"
    REMITTITUR = "060remittitur"
    REHEARING = "070rehearing"
    ON_THE_MERITS = "080onthemerits"
    ON_MOTION_TO_STRIKE_COST_BILL = "090onmotiontostrike"



"""Scraper for Armed Services Board of Contract Appeals
CourtID: asbca
Court Short Name: ASBCA
Author: Jon Andersen
Reviewer: mlr
History:
    2014-09-11: Created by Jon Andersen
    2016-03-17: Website and phone are dead. Scraper disabled in __init__.py.
"""

from datetime import datetime

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.year = str(datetime.today().year)
        self.url = f"https://www.asbca.mil/Decisions/decisions{self.year}.html"
        self.status = "Published"
        self.request["headers"] = {
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "en-US,en;q=0.9",
            "priority": "u=0, i",
            "sec-ch-ua": '"Chromium";v="134", "Not:A-Brand";v="24", "Google Chrome";v="134"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Linux"',
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "none",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
            "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36",
        }
        self.needs_special_headers = True

    def _process_html(self):
        # Exclude headers and rows that only have the month name
        if self.test_mode_enabled():
            self.year = "2024"

        rows = self.html.xpath(
            "//tr[not(th) and not(.//span[@style='background-color:#F8C100;'])]"
        )
        for row in rows:
            if len(row.xpath(".//td")) != 4:
                logger.warning(
                    "Row does not have expected number of cells %s",
                    row.text_content().strip(),
                )
                continue

            url = row.xpath(".//a/@href")
            url = url[0]
            date, docket, name, judge = (
                cell.text_content().strip() for cell in row.xpath(".//td")
            )

            if self.year not in date:
                # site returns all records in a single request
                # in a normal scrape, check only the most recent year
                break

            self.cases.append(
                {
                    "date": date,
                    "name": name,
                    "url": url,
                    "docket": docket,
                    "judge": judge,
                }
            )



# Scraper for Texas 4th Court of Appeals
# CourtID: texapp4
# Court Short Name: TX
# Author: Andrei Chelaru
# Reviewer: mlr
# Date: 2014-07-10


from juriscraper.opinions.united_states.state import texapp


class Site(texapp.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_name = "capp_4"
        self.checkbox = 5



# Author: William E. Palin
# Date created: 2023-01-04
# Notes: See Alabama

from juriscraper.opinions.united_states.state import ala


class Site(ala.Site):
    court_str = "b82b30d5-bd3c-46d7-9451-1cb05e470873"



# History:
# - Long ago: Created by mlr
# - 2014-11-07: Updated by mlr to use new website.
# - 2025-08-26: Updated by lmanzur to use OpinionSiteLinear and extract lower court

import re

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "http://www.ca5.uscourts.gov/rss.aspx?Feed=Opinions&Which=All&Style=Detail"
        self.court_id = self.__module__

    def _process_html(self):
        for item in self.html.xpath("//item"):
            case_name = item.xpath("description/text()[2]")[0]
            download_url = item.xpath("link")[0].tail
            case_date = item.xpath("description/text()[5]")[0]
            docket_number = item.xpath("description/text()[1]")[0]
            status_raw = item.xpath("description/text()[3]")[0]
            nature_of_suit = item.xpath("description/text()[4]")[0]

            if status_raw == "pub":
                status = "Published"
            elif status_raw == "unpub":
                status = "Unpublished"
            else:
                status = "Unknown"

            # Ensure date is a string, and clean up if needed
            case_date_cleaned = case_date.strip()

            self.cases.append(
                {
                    "name": case_name,
                    "url": download_url,
                    "date": case_date_cleaned,  # Now always a string
                    "docket": docket_number,
                    "status": status,
                    "nature_of_suit": nature_of_suit,
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """

        pattern = re.compile(
            r"""
            (?:
               Appeals?\s+from\s+the\s+
             | Petition\s+for\s+Review\s+from\s+an\s+Order\s+of\s+the\s+
             | Petition\s+for\s+Review\s+of\s+the\s+
            )
            (?P<lower_court>[^.]+?)
            (?=\s*(?:\.|Nos?\.|USDC))
            """,
            re.X,
        )

        lower_court = ""
        if match := pattern.search(scraped_text):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()

        if lower_court:
            return {
                "Docket": {
                    "appeal_from_str": lower_court,
                }
            }
        return {}



import re
from datetime import date, datetime, timedelta
from urllib.parse import urlencode, urljoin

from dateutil import parser
from requests.exceptions import ChunkedEncodingError

from juriscraper.AbstractSite import logger
from juriscraper.lib.html_utils import (
    get_row_column_text,
)
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    # Sharin S. Anderson v Alyeska Pipeline Service Co.; Opinion Number: 6496
    first_opinion_date = datetime(2010, 7, 23).date()
    base_url = "https://appellate-records.courts.alaska.gov/CMSPublic/"
    # endpoint for opinion url without parameters
    op_endpoint = urljoin(base_url, "UserControl/OpenOpinionDocument")
    is_coa = False
    search_parameter = "Opinions"
    document_type = "OP"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_backscrape = False
        self.make_backscrape_iterable(kwargs)
        self.court_id = self.__module__
        self.url = urljoin(
            self.base_url, f"Home/{self.search_parameter}?isCOA={self.is_coa}"
        )
        self.status = "Published"
        # juriscraper in the user agent crashes it
        # it appears to be just straight up blocked.
        self.request["headers"]["user-agent"] = "Free Law Project"
        self.end_date = date.today()
        self.start_date = self.end_date - timedelta(days=30)
        self.is_citation_visible = True

    def _download(self, request_dict=None):
        # Unfortunately, about 2/3 of calls are rejected by alaska but
        # if we just ignore those encoding errors we can live with it
        if request_dict is None:
            request_dict = {}
        try:
            return super()._download(request_dict)
        except ChunkedEncodingError:
            return None

    def _process_html(self) -> None:
        if not self.html:
            logger.info(
                "HTML was not downloaded from source page. Should retry"
            )
            return
        for table in self.html.xpath("//table"):
            adate = table.xpath("./preceding-sibling::h5")[0].text_content()
            if (
                not self.date_is_in_range(adate)
                and not self.test_mode_enabled()
            ):
                if self.is_backscrape:
                    logger.debug("Backscraper skipping %s", adate)
                    continue
                break

            for row in table.xpath(".//tr"):
                if row.text_content().strip():
                    case_number = get_row_column_text(row, 3)
                    name = get_row_column_text(row, 4)
                    citation = (
                        get_row_column_text(row, 5)
                        if self.is_citation_visible
                        else ""
                    )

                    # get parameters required to build opinion url
                    params = {
                        "docNumber": get_row_column_text(row, 2),
                        "caseNumber": case_number.split(",")[0],
                        "opinionType": self.document_type,
                    }
                    url = f"{self.op_endpoint}?{urlencode(params)}"

                    self.cases.append(
                        {
                            "date": adate,
                            "docket": case_number,
                            "name": name,
                            "citation": citation,
                            "url": url,
                        }
                    )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """
        pattern = re.compile(
            r"""
            (?:
                Appeals?\s+from\s+the\s+|
                Appeals\s+in\s+File\s+Nos\.\s+S-\d+(?:/\d+)?\s+from\s+the\s+|
                Petitions?\s+for\s+Review\s+from\s+the\s+|
                Petition\s+for\s+Hearing\s+from\s+the\s+|
                Certified\s+(?:Original\s+Application\s+for\s+Relief\s+and\s+Jurisdiction\s+Transfer|Question)\s+from\s+the\s+
            )
            (?P<lower_court>.*?)(?=Judge\.|\n\s*\n)
            """,
            re.X | re.DOTALL,
        )

        lower_court_str = ""
        if match := pattern.search(scraped_text):
            lower_court_str = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()

        if lower_court_str:
            parts = [part.strip() for part in lower_court_str.split(",")]
            if "appeals" in parts[0].lower():
                lower_court = parts[0]
            else:
                lower_court = ", ".join(parts[:2])
            # Handle judge name, including cases like "Jr."
            lower_court_judge = None
            if len(parts) > 1:
                lower_court_judge = parts[-2]
                if lower_court_judge == "Jr." and len(parts) > 2:
                    lower_court_judge = f"{parts[-3]} Jr."

            return_dict = {
                "Docket": {
                    "appeal_from_str": lower_court,
                },
            }
            if lower_court_judge:
                return_dict.setdefault("OriginatingCourtInformation", {})[
                    "assigned_to_str"
                ] = lower_court_judge

            return return_dict

        return {}

    def make_backscrape_iterable(self, kwargs: dict) -> None:
        """Checks if backscrape start and end arguments have been passed
        by caller, and parses them accordingly

        Alaska's opinions page returns all opinions, so we must only load it
        (successfully, because it may load partially) once.
        We can use the backscrape arguments to filter out opinions not in the
        date range we are looking for

        :return None
        """
        start = kwargs.get("backscrape_start")
        end = kwargs.get("backscrape_end")

        if start:
            start = datetime.strptime(start, "%m/%d/%Y").date()
        else:
            start = self.first_opinion_date
        if end:
            end = datetime.strptime(end, "%m/%d/%Y").date()
        else:
            end = datetime.now().date()

        self.back_scrape_iterable = [(start, end)]

    def _download_backwards(self, dates: tuple[date]) -> None:
        """Called when backscraping

        :param dates: (start_date, end_date) tuple
        :return None
        """
        self.start_date, self.end_date = dates
        self.is_backscrape = True
        logger.info(
            "Backscraping for range %s %s", self.start_date, self.end_date
        )

    def date_is_in_range(self, date_str: str) -> bool:
        """Check if the table date is in
        the range

        :param date_str: string date from the HTML source
        :return: True if date is in backscrape range
        """
        parsed_date = parser.parse(date_str).date()
        return self.start_date <= parsed_date <= self.end_date



from datetime import datetime

from juriscraper.opinions.united_states_backscrapers.state import mass


class Site(mass.Site):
    docket_number_regex = r"\d{2}-\w-\d{1,4}"
    backscrape_date_range_mapper = [
        {
            "start": datetime(2021, 7, 9),
            "end": None,
            "url": "http://masscases.com/app100-124.html",
        },
        {
            "start": datetime(2009, 8, 20),
            "end": datetime(2021, 5, 12),
            "url": "http://masscases.com/app75-99.html",
        },
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__



# Scraper for California's Superior Court Appellate Division
# CourtID: calctapp_app_div
# Court Short Name: Cal. Ct. App.

from juriscraper.opinions.united_states.state import calctapp_1st


class Site(calctapp_1st.Site):
    court_code = "J"
    division = "App. Div."



# Scraper for Minnesota Supreme Court
# CourtID: minn
# Court Short Name: MN
# Author: Andrei Chelaru
# Reviewer: mlr
# Date: 2014-07-03
# Contact:  Liz Reppe (Liz.Reppe@courts.state.mn.us), Jay Achenbach (jay.achenbach@state.mn.us)

# 2024-08-09: update and implement backscraper, grossir

import re
from datetime import date
from urllib.parse import urljoin

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    court_query = "supct"
    days_interval = 7
    first_opinion_date = date(1998, 1, 1)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__

        self.status = "Published"
        if self.court_query == "ctapun":
            self.status = "Unpublished"

        self.url = "https://mn.gov/law-library/search/"
        self.params = self.base_params = {
            "v:sources": "mn-law-library-opinions",
            "query": f" (url:/archive/{self.court_query}) ",
            "sortby": "date",
        }
        self.request["verify"] = False
        self.request["headers"] = {
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Fetch-Dest": "document",
            "Accept-Language": "en-US,en;q=0.9",
            "Sec-Fetch-Mode": "navigate",
            "Host": "mn.gov",
            "User-Agent": "Juriscraper/3.0 Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.1 Safari/605.1.15",
            "Referer": "https://mn.gov/law-library/search/?v%3Asources=mn-law-library-opinions&query=+%28url%3A%2Farchive%2Fsupct%29+&citation=&qt=&sortby=&docket=&case=&v=&p=&start-date=&end-date=",
            "Connection": "keep-alive",
        }
        self.make_backscrape_iterable(kwargs)
        self.needs_special_headers = True

    def _process_html(self) -> None:
        """Process the html and extract out the opinions

        :return: None
        """
        self.html = self._download({"params": self.params})

        # This warning is useful for backscraping
        results_number = self.html.xpath(
            "//div[@class='searchresult_number']/text()"
        )
        if results_number:
            results_number = results_number[0].split(" of ")[-1].strip()
            if int(results_number) > 10:
                logger.warning(
                    "Page has a size 10, and there are %s results for this query",
                    results_number,
                )

        for case in self.html.xpath("//div[@class='searchresult']"):
            name = re.sub(r"\s+", " ", case.xpath(".//a/text()")[0])
            name_match = re.search(r"(?P<name>.+)\. A\d{2}-\d+", name)
            if name_match:
                name = name_match.group("name")

            summary = case.xpath(
                ".//div[@class='searchresult_snippet']/text()"
            )[0]
            disposition = ""
            # minnctapp_u sometimes has a disposition, but it's format
            # is much more variable
            if self.status == "Published":
                parts = summary.split(".")
                if (
                    len(parts) > 2
                    and len(parts[-2].split()) <= 10
                    and not re.search(r"\d", parts[-2])
                ):
                    # Sometimes there is no disposition. Ex: A23-1504
                    # False positives usually contain numbers. Ex: A22-0776
                    disposition = parts[-2].strip()

            raw_date = case.xpath(".//div[@class='searchresult_date']")[
                -1
            ].text_content()
            date_filed = re.sub(r"\s+", " ", raw_date).split(":")[1].strip()

            # Backscrapers may find citations, since they are attached some
            # months after the initial release of an opinion
            citation = ""
            if cite_element := case.xpath(
                ".//div[b[contains(text(), 'Citation:')]]"
            ):
                citation = (
                    cite_element[0].text_content().split(":", 1)[-1].strip()
                )

            url = case.xpath(".//a/@href")[0]
            docket = url.split("/")[-1].split("-")[0][2:]
            self.cases.append(
                {
                    "date": date_filed,
                    "name": name,
                    "url": urljoin("https://", url),
                    "disposition": disposition,
                    "summary": summary,
                    "docket": docket,
                    "citation": citation,
                }
            )

    def _download_backwards(self, dates: tuple[date]):
        logger.info("Backscraping for range %s - %s", *dates)
        params = {**self.base_params}
        params.update(
            {
                "start-date": dates[0].strftime("%-m/%-d/%Y"),
                "end-date": dates[1].strftime("%-m/%-d/%Y"),
                "query": f"{params['query']}date:[{dates[0].strftime('%Y-%m-%d')}..{dates[1].strftime('%Y-%m-%d')}]",
            }
        )
        self.params = params



from juriscraper.opinions.united_states.federal_appellate import cadc


class Site(cadc.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # https://media.cadc.uscourts.gov/judgments/
        self.url = "https://media.cadc.uscourts.gov/judgments/bydate/recent"
        self.status = "Unpublished"



"""
Scraper for Kansas Court of Appeals
CourtID: kanctapp
Court Short Name: Kansas Ct. App.
Author: William Palin
Court Contact:
History:
 - 2023-01-04: Created.
Notes:
 - selenium
"""

from juriscraper.opinions.united_states.state import kan


class Site(kan.Site):
    link_xp = '//div[@class="col-md-6"][2]/ul/li'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__



# Author: Michael Lissner
# Date created: 2013-05-23

import re
from datetime import date, datetime

from juriscraper.lib.date_utils import unique_year_month
from juriscraper.lib.log_tools import make_default_logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear

logger = make_default_logger()


class Site(OpinionSiteLinear):
    first_opinion_date = datetime(2010, 1, 1)
    days_interval = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = (
            "https://www.courts.state.hi.us/opinions_and_orders/opinions"
        )
        self.court_id = self.__module__
        self.court_code = "S.Ct"
        self.status = "Published"
        self.should_have_results = True
        self.make_backscrape_iterable(kwargs)

    def _process_html(self) -> None:
        """Parse HTML into case objects

        :return: None
        """
        for row in self.html.xpath("//tr[@class='row-']"):
            date, court, docket, name, lower_court, citation = row.xpath(
                ".//td"
            )
            court = court.text_content()
            if court != self.court_code:
                continue

            if not docket.xpath(".//a"):
                continue

            name_text = name.text_content().strip()
            last_period_regex = (
                r", Dated|(?<!v)(?<!Jr)(?<!U\.S)(?<! [A-Z])\.(?! Inc\.)"
            )

            if "(" in name_text:
                name = name_text.split("(", 1)[0].strip()
            else:
                v_match = re.search(r" v\. ", name_text)
                if v_match:
                    post_v = name_text[v_match.end() :]
                    name = (
                        name_text[: v_match.end()]
                        + re.split(last_period_regex, post_v, maxsplit=1)[
                            0
                        ].strip()
                    )
                else:
                    name = re.split(last_period_regex, name_text, maxsplit=1)[
                        0
                    ].strip()

            self.cases.append(
                {
                    "date": date.text_content(),
                    "name": name,
                    "docket": docket.text_content()
                    .strip()
                    .split("\t")[0]
                    .split()[0],
                    "url": docket.xpath(".//a")[0].get("href"),
                    "lower_court": lower_court.text_content(),
                    "citation": citation.text_content(),
                }
            )

    def _download_backwards(self, search_date: date) -> None:
        """Download and process HTML for a given target date.

        :param search_date (date): The date for which to download and process opinions.
        :return None; sets the target date, downloads the corresponding HTML
        and processes the HTML to extract case details.
        """

        self.request["parameters"]["params"] = {
            "yr": search_date.year,
            "mo": f"{search_date.month:02d}",
        }
        logger.info(
            "Now downloading case page at: %s (params: %s)"
            % (self.url, self.request["parameters"]["params"])
        )
        self.html = self._download()
        self._process_html()

    def make_backscrape_iterable(self, kwargs) -> None:
        """Make back scrape iterable

        :param kwargs: the back scraping params
        :return: None
        """
        super().make_backscrape_iterable(kwargs)
        self.back_scrape_iterable = unique_year_month(
            self.back_scrape_iterable
        )



"""
Scraper for New Hampshire Supreme Court
CourtID: nh_p
Court Short Name: NH
Court Contact: webmaster@courts.state.nh.us
Author: Andrei Chelaru
Reviewer: mlr
History:
    - 2014-06-27: Created
    - 2014-10-17: Updated by mlr to fix regex error.
    - 2015-06-04: Updated by bwc so regex catches comma, period, or
    whitespaces as separator. Simplified by mlr to make regexes more semantic.
    - 2016-02-20: Updated by arderyp to handle strange format where multiple
    case names and docket numbers appear in anchor text for a single case pdf
    link. Multiple case names are concatenated, and docket numbers are
    concatenated with ',' delimiter
    - 2021-12-29: Updated for new web site, by flooie and satsuki-chan
    - 2024-08-21: Implement backscraper and update headers, by grossir
"""

import re
from datetime import datetime
from urllib.parse import urlencode, urljoin

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    """
    Regular site:
    https://www.courts.nh.gov/our-courts/supreme-court/orders-and-opinions/opinions
    """

    # document_purpose = 1331 -> Supreme Court Opinion
    base_filter = "{}@field_document_purpose|=|1331"
    year_to_filter = {
        2024: "@field_document_subcategory|=|2316",
        2023: "@field_document_subcategory|=|2256",
        2022: "@field_document_subcategory|=|2091",
        2021: "@field_tags|CONTAINS|1206~field_entity_tags|CONTAINS|1206",
        2020: "@field_tags|CONTAINS|1366~field_entity_tags|CONTAINS|1366",
        2019: "@field_tags|CONTAINS|1416~field_entity_tags|CONTAINS|1416",
        2018: "@field_document_subcategory|=|1601",
        2017: "@field_document_subcategory|=|1596",
        2016: "@field_document_subcategory|=|1591",
        2015: "@field_document_subcategory|=|1586",
        2014: "@field_document_subcategory|=|1581",
    }
    filter_mode = "exclusive"
    document_type = "opinions"
    # there is data since 2002, but we would need to
    # collect all subcategory or tag values
    start_year = 2015
    end_year = datetime.today().year - 1
    cite_regex = re.compile(r"\d{4}\sN\.H\.\s\d+")
    docket_regex = re.compile(r"(?P<docket>\d{4}-\d{1,4})[\s,]*")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.status = "Published"
        self.base_url = "https://www.courts.nh.gov/content/api/documents"
        self.set_request_parameters()
        self.needs_special_headers = True
        self.make_backscrape_iterable(kwargs)
        self.paginate = False

    def _process_html(self) -> None:
        json_response = self.html

        for case in json_response["data"]:
            fields = case["fields"]

            if fields["field_document_file"]:
                url_object = fields["field_document_file"]
            elif fields["field_document"]:
                # seen on withdrawn opinions. Ex: 2020-0268
                # Hampstead School Board & a. v. School Administrative Unit No. 5
                url_object = fields["field_document"][0]["fields"][
                    "field_document_file"
                ]
            else:
                logger.warning(
                    "Skipping row '%s', can't find document link",
                    case["title"],
                )
                continue
            url = url_object["0"]["fields"]["uri"][0].split("//")[1]

            if fields["field_date_filed"]:
                case_date = fields["field_date_filed"][0]
            elif fields["field_date_posted"]:
                # usually this is the only populated field
                case_date = fields["field_date_posted"][0]
            else:
                logger.warning(
                    "Skipping row '%s'. No date found", case["title"]
                )
                continue

            # "title" format changes since 2024, where a citation string
            # replaces the docket number, and the docket is in another row
            name = case["title"]

            citation = ""
            cite_match = self.cite_regex.search(case["title"])
            if cite_match:
                citation = cite_match.group(0)
                name = re.sub(self.cite_regex, " ", name)

            docket = ", ".join(
                [
                    match.group("docket")
                    for match in self.docket_regex.finditer(case["title"])
                ]
            )
            if not docket and fields.get("field_description"):
                docket_str = fields["field_description"][0]["#text"]
                docket = self.docket_regex.search(docket_str).group(0)

            name = re.sub(self.docket_regex, " ", name)
            # delete traces of multiple docket numbers
            name = re.sub(r"^(and|[&,])", "", name.strip()).strip()

            self.cases.append(
                {
                    "date": case_date,
                    "url": urljoin(
                        "https://www.courts.nh.gov/sites/g/files/ehbemt471/files/",
                        url,
                    ),
                    "name": name,
                    "docket": docket,
                    "citation": citation,
                }
            )

        # This flag will be set to True by the _download_backwards method
        if not self.paginate:
            return
        self.paginate = False  # prevent recursion

        logger.info(
            "Found %s results, will paginate through", json_response["total"]
        )
        for page in range(2, json_response["last_page"] + 1):
            logger.info("Paginating to page %s", page)
            self.url = self.url.replace(f"page={page - 1}", f"page={page}")
            self.html = self._download()
            self._process_html()

    def set_request_parameters(
        self, year: int = datetime.today().year
    ) -> None:
        """Each year has a unique `field_document_subcategory` key, so we must
        set it accordingly

        :param year: full year integer
        """
        year_filter = (
            "" if not self.back_scrape_iterable else self.year_to_filter[year]
        )
        params = {
            "iterate_nodes": "true",
            # Will raise a KeyError if there is no proper year key, we will
            # need to manually correct this next year
            "q": self.base_filter.format(year_filter),
            "sort": "field_date_posted|desc|ALLOW_NULLS",
            "filter_mode": self.filter_mode,
            "type": "document",
            "page": "1",
            "size": "25",
        }
        self.url = f"{self.base_url}?{urlencode(params)}"
        self.request["headers"] = {
            "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36",
            "Sec-Ch-Ua": '"Google Chrome";v="135", "Not-A.Brand";v="8", "Chromium";v="135"',
            "Referer": f"https://www.courts.nh.gov/our-courts/supreme-court/orders-and-opinions/{self.document_type}/{year}",
            "X-Requested-With": "XMLHttpRequest",
        }

    def _download_backwards(self, year: int) -> None:
        self.paginate = True
        self.set_request_parameters(year)
        logger.info("Backscraping year %s", year)
        self.html = self._download()
        self._process_html()

    def make_backscrape_iterable(self, kwargs: dict) -> list[int]:
        """The API exposes no date filter, so we must query a year
        and then paginate the results.
        """
        start = int(kwargs.get("backscrape_start") or self.start_year)
        end = int(kwargs.get("backscrape_end") or self.end_year)

        if start == end:
            self.back_scrape_iterable = [start]
        else:
            self.back_scrape_iterable = range(start, end + 1)



import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Optional

from lxml import html
from lxml.html import HtmlElement

from juriscraper.AbstractSite import logger
from juriscraper.lib.html_utils import clean_html, strip_bad_html_tags_insecure
from juriscraper.lib.string_utils import clean_string, harmonize
from juriscraper.scotus import SCOTUSDocketReport


@dataclass
class ContactAddress:
    title: Optional[str]
    address_lines: list[str]
    city: Optional[str]
    state: Optional[str]
    zip_code: Optional[str]
    email: Optional[str]


class SCOTUSDocketReportHTML(SCOTUSDocketReport):
    """Parse SCOTUS docket HTML."""

    EMAIL_RE = re.compile(r"[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,}", re.I)
    ID_RE = re.compile(r"#\s*[A-Za-z0-9-]+\b")
    ADDRESS_NUMBER = re.compile(
        r"\b(\d{1,6}(?:-\d{1,6})?(?:\s+\d+\/\d+)?[A-Za-z]?)\b"
    )
    ADDRESS_RE = r"([^,]+),\s*([A-Z]{2})\s+(\d{5}(?:-\d{4})?)$"
    DOCKET_NUMBER_RE = r"No\.\s*([^\s<]+)"

    def __init__(self, court_id: str = "scotus"):
        """Initialize the HTML report parser."""
        super().__init__(court_id=court_id)
        self.tree = None

    def _parse_text(self, text: str) -> None:
        """Parse raw HTML and store a lxml tree.

        :param text: The raw HTML unicode object.
        :return: None
        """
        text = clean_html(text)
        self.tree = html.fromstring(text or "")

    @property
    def metadata(self) -> dict[str, Any]:
        """Return normalized docket metadata extracted from HTML.

        :return: A dict with the SCOTUS docket metadata.
        """

        if self.tree is None:
            return {}

        docket_title_text = self._join_nodes_content(
            self.tree.xpath(
                "//table[@id='docketinfo']//span[contains(@class,'DocketInfoTitle')]"
            )
        )
        docket_number = None
        if docket_title_text:
            m = re.search(self.DOCKET_NUMBER_RE, docket_title_text)
            docket_number = m.group(1).strip() if m else None

        case_name = clean_string(
            harmonize(
                self._clean_whitespace(
                    self._join_nodes_content(
                        self.tree.xpath(
                            "//table[@id='docketinfo']//span[@class='title']"
                        )
                    )
                )
            )
        )

        docketed_str = self._td_value_by_label("Docketed:")
        date_filed = self.normalize_date(docketed_str)

        linked_with_val = self._td_value_by_label("Linked with:")
        if not linked_with_val:
            # Sometimes the value is in the same cell: Linked with 16A827
            target_label = "string(//table[@id='docketinfo']//tr[td[1]/span[starts-with(normalize-space(.), 'Linked with')]])"
            linked_with_val = self.tree.xpath(target_label).replace(
                "Linked with", ""
            )
        links = (linked_with_val or "").strip()

        lower_court = self._td_value_by_label("Lower Ct:")
        # Lower court case numbers
        lower_court_raw = self._td_value_by_label(
            "Case Numbers:", allow_indent=True
        )
        lower_court_raw = (lower_court_raw or "").strip()
        lower_court_cleaned = None
        if lower_court_raw:
            lower_court_text = lower_court_raw.strip("()")
            lower_court_cleaned = self.clean_lower_court_cases(
                lower_court_text
            )

        # Decision dates
        lower_court_decision_date = self.normalize_date(
            self._td_value_by_label("Decision Date:", allow_indent=True)
        )
        lower_court_rehearing_denied = self.normalize_date(
            self._td_value_by_label("Rehearing Denied:", allow_indent=True)
        )
        discretionary_court_decision = self.normalize_date(
            self._td_value_by_label(
                "Discretionary Court Decision Date:", allow_indent=True
            )
        )

        # Questions Presented
        questions_presented_label = self.tree.xpath(
            "//tr[td[1][starts-with(normalize-space(.), 'Questions Presented')]]//a[@href]"
        )
        questions_presented_val = (
            questions_presented_label[0].get("href")
            if questions_presented_label
            else None
        )

        return {
            "docket_number": docket_number,
            "capital_case": None,
            "date_filed": date_filed,
            "case_name": case_name,
            "links": links,
            "lower_court": lower_court or None,
            "lower_court_case_numbers_raw": lower_court_raw or None,
            "lower_court_case_numbers": lower_court_cleaned,
            "lower_court_decision_date": lower_court_decision_date,
            "lower_court_rehearing_denied_date": lower_court_rehearing_denied,
            "questions_presented": questions_presented_val or None,
            "discretionary_court_decision": discretionary_court_decision,
        }

    @staticmethod
    def _parse_description_html(td) -> str:
        """Parse the cell content up to the first <br>, excluding
        .documentlinks.

        :param td: <td> element with the text description.
        :return: The description_html.
        """
        parts = []
        if td is None:
            return ""

        if td.text:
            parts.append(td.text)

        for child in td:
            tag = getattr(child, "tag", "")
            if tag and tag.lower() == "br":
                break
            # If a documentlinks span appears before a <br>, stop there too.
            if "documentlinks" in (child.get("class") or ""):
                break

            parts.append(html.tostring(child, encoding="unicode"))
            # Include its tail too.
            if child.tail:
                parts.append(child.tail)

        return "".join(parts)

    def _build_docket_entry(
        self,
        date_str: str,
        description_td: Optional[HtmlElement],
        attachment_path: str,
        table_layout: bool = False,
    ) -> dict[str, Any]:
        """Build a normalized docket entry dict.

        :param date_str: Raw date string from the date cell.
        :param description_td: The <td> element containing the description.
        :param attachment_path: XPath to select attachment anchors inside the
        description cell.
        :param table_layout: If parsing a table layout docket.
        :return: Dict with date_filed, description, description_html, attachments.
        """
        description_html = ""
        attachments: list[dict[str, Any]] = []

        if description_td is not None:
            # Select content up to first <br> and excluding .documentlinks;
            # fallback to whole cell.
            description_html = (
                self._parse_description_html(description_td).strip()
                or html.tostring(description_td, encoding="unicode").strip()
            )
            # Parse attachments.
            for a in description_td.xpath(attachment_path):
                href = a.get("href")
                if not href:
                    continue

                if table_layout and href:
                    logger.error(
                        "Found a potential attachment in a table-layout docket. "
                        "This could be an opportunity to add support for it."
                    )

                short_desc = self._clean_whitespace(
                    (a.text or a.get("title") or "").strip()
                )
                attachments.append(
                    {
                        "short_description": short_desc or "Document",
                        "document_url": href,
                    }
                )

        cleaned = strip_bad_html_tags_insecure(description_html or "")
        description = self._clean_whitespace(cleaned.text_content())

        return {
            "date_filed": self.normalize_date(date_str),
            "description": description,
            "description_html": description_html,
            "attachments": attachments,
        }

    @property
    def docket_entries(self) -> list[dict[str, Any]]:
        """Return docket entries from 'Proceedings and Orders'.

        :return: List of dicts with date_filed, description, description_html,
        attachments.
        """
        if self.tree is None:
            return []

        # Div-based layout
        entries: list[dict[str, Any]] = []
        rows = self.tree.xpath(
            "//div[@id='proceedings']//div[contains(@class,'card')]"
            "//table[contains(@class,'ProceedingItem')]//tr"
        )
        for tr in rows:
            date_td = tr.xpath("./td[contains(@class,'ProceedingDate')]")
            text_td = tr.xpath("./td[not(contains(@class,'ProceedingDate'))]")
            date_str = self._join_nodes_content(date_td)
            description_td = text_td[0] if text_td else None
            entries.append(
                self._build_docket_entry(
                    date_str=date_str,
                    description_td=description_td,
                    attachment_path=".//span[contains(@class,'documentlinks')]//a[contains(@class,'documentanchor')]",
                )
            )
        if entries:
            return entries

        # Table-based layout
        table = self.tree.xpath("//table[@id='proceedings']")
        if not table:
            return []

        entries = []
        for tr in table[0].xpath(".//tr[td]"):
            # omit proceedingheader header
            if tr.xpath(
                "./td[contains(concat(' ', normalize-space(@class), ' '), ' proceedingheader ')]"
            ):
                continue
            # omit borderbttm separator
            if tr.xpath(
                "./td[contains(concat(' ', normalize-space(@class), ' '), ' borderbttm ')]"
            ):
                continue

            # omit non entries cells.
            tds = tr.xpath("./td")
            if len(tds) < 2:
                continue

            date_td, text_td = tds[:2]
            date_str = self._clean_whitespace(date_td.text_content())

            # Table layout, any link in the cell is a potential attachment.
            entries.append(
                self._build_docket_entry(
                    date_str=date_str,
                    description_td=text_td,
                    attachment_path=".//a[@href]",
                    table_layout=True,
                )
            )

        return entries

    @property
    def parties(self) -> list[dict[str, Any]]:
        """Return parties grouped under Contacts (Petitioner/Respondent/Other).

        Tries the card-based layout first. If not found, tries the table-based layout.
        :return: List of dicts containing parties data.
        """
        if self.tree is None:
            return []

        parties = self._parties_from_cards()
        if parties:
            return parties

        return self._parties_from_contacts_table()

    def _parties_from_cards(self) -> list[dict[str, Any]]:
        """Parse parties from the Contacts card-based layout.

        :return: List of dicts containing parties data.
        """
        sections = [
            ("Attorneys for Petitioners", "Petitioner"),
            ("Attorneys for Respondents", "Respondent"),
            ("Other Attorneys", "Other"),
        ]

        parties = []

        for heading_text, type_key in sections:
            section_root = self._section_by_heading(heading_text)
            if section_root is None:
                continue

            names = section_root.xpath(
                ".//div[contains(@class,'ContactName')]"
            )
            contact_data = section_root.xpath(
                ".//div[contains(@class,'ContactData2')]"
            )

            attorneys_by_party = defaultdict(list)
            for i, name_div in enumerate(names):
                data_div = contact_data[i] if i < len(contact_data) else None
                attorney = self._build_attorney_from_contact(
                    name_div, data_div
                )
                # Normalize party_name
                party_name = ""
                if data_div is not None:
                    party_name_raw = data_div.xpath(
                        ".//span[contains(@class,'partyname')]/text()"
                    )
                    party_name = (
                        self._clean_whitespace("".join(party_name_raw)) or ""
                    )
                party_name = party_name.replace("Party name:", "").strip()
                attorneys_by_party[(type_key, party_name)].append(attorney)

            parties.extend(
                {"type": t, "name": n, "attorneys": attys}
                for (t, n), attys in attorneys_by_party.items()
            )
        return parties

    def _parties_from_contacts_table(self) -> list[dict[str, Any]]:
        """Parse parties from the table-based contacts layout.

        :return: List of dicts containing parties data.
        """
        table = next(iter(self.tree.xpath("//table[@id='Contacts']")), None)
        if table is None:
            return []

        parties_by_key: dict[tuple[str, str], list] = defaultdict(list)
        current_type = None
        rows = table.xpath(".//tr[td]")
        for i, tr in enumerate(rows):
            # omit section header
            if tr.xpath(
                "./td[contains(concat(' ', normalize-space(@class), ' '), ' ContactSubHeader ')]"
            ):
                header_text = self._clean_whitespace(
                    "".join(
                        tr.xpath(
                            ".//span[contains(@class,'tableheadertext')]/text()"
                        )
                    )
                )
                current_type = self._map_contacts_header_to_type(header_text)
                continue

            # Attorney row
            tds = tr.xpath("./td")
            if current_type and len(tds) == 3:
                name_td, addr_td, phone_td = tds
                name_text = self._clean_whitespace(
                    self._parse_description_html(name_td) or ""
                )
                full_name_cell = self._clean_whitespace(name_td.text_content())
                is_counsel_of_record = bool(
                    re.search(r"\bcounsel of record\b", full_name_cell, re.I)
                )

                contact_data = self._parse_contacts_table_address_cell(addr_td)
                title = contact_data.title
                address_lines = contact_data.address_lines
                city = contact_data.city
                state = contact_data.state
                zip_code = contact_data.zip_code
                email = contact_data.email
                phone = self._clean_whitespace(phone_td.text_content()) or None

                # Party row
                party_name = ""
                if i + 1 < len(rows):
                    next_tr = rows[i + 1]
                    if next_tr.xpath(
                        "./td[contains(concat(' ', normalize-space(@class), ' '), ' ContactParty ')]"
                        " | ./td[contains(normalize-space(string(.)), 'Party name:')]"
                    ):
                        raw_party = self._clean_whitespace(
                            next_tr.text_content()
                        )
                        party_name = re.sub(
                            r"^Party name:\s*", "", raw_party, flags=re.I
                        ).strip()

                # Build attorney
                attorney = {
                    "name": name_text or None,
                    "is_counsel_of_record": is_counsel_of_record,
                    "title": title,
                    "phone": phone,
                    "address": ", ".join(address_lines)
                    if address_lines
                    else None,
                    "city": city,
                    "state": state,
                    "zip": zip_code,
                    "email": email,
                }
                # Group parties by type-name
                parties_by_key[(current_type, party_name or "")].append(
                    attorney
                )

        return [
            {"type": party_type, "name": name, "attorneys": attorneys}
            for (party_type, name), attorneys in parties_by_key.items()
        ]

    @staticmethod
    def _map_contacts_header_to_type(header_text: str) -> Optional[str]:
        """Map table section headers to normalized types.

        :param header_text: Text to be mapped.
        :return: Normalized party type or None.
        """
        t = (header_text or "").strip().lower()
        if "petitioner" in t:
            return "Petitioner"
        if "respondent" in t:
            return "Respondent"
        if "other" in t:
            return "Other"
        return None

    def _append_clean_text(
        self, text: Optional[str], lines: list[str]
    ) -> None:
        """Clean the given text and append it to the list of lines if not empty."""
        cleaned = self._clean_whitespace(text)
        if cleaned:
            lines.append(cleaned)

    def _parse_address_title(self, lines) -> tuple[Optional[str], int]:
        """Extract the party title from address lines and determine where the address begins.

        :param lines: A list of text lines containing the party title and address.
        :return: A twp tuple where the first element is the concatenated title
        string and th index where the address starts.
        """
        title_parts = []
        start_add_idx = 0
        # Parse title.
        for j, ln in enumerate(lines):
            if self.ID_RE.match(ln):
                # Omit lines starting with # like #1098260
                continue
            if self.ADDRESS_NUMBER.search(ln):
                # Match the start of an address by looking for a street number.
                start_add_idx = j
                break
            title_parts.append(ln)

        title = ", ".join(title_parts) or None
        return title, start_add_idx

    def _parse_address_lines(self, lines, start_add_idx) -> ContactAddress:
        """Parse the address components starting from a given index.

        :param lines: A list of text lines containing the address data.
        :param start_add_idx: Index indicating where the address lines begin.
        :return: A ContactAddress object.
        """

        address_lines = (
            lines[start_add_idx:] if start_add_idx < len(lines) else []
        )

        # parse City/State/Zip
        city = state = zip_code = None
        if address_lines:
            last = address_lines[-1]
            if m := re.search(self.ADDRESS_RE, last):
                city, state, zip_code = (
                    m.group(1).strip(),
                    m.group(2),
                    m.group(3),
                )
                address_lines = address_lines[:-1]

        return ContactAddress(
            title=None,
            address_lines=address_lines,
            city=city,
            state=state,
            zip_code=zip_code,
            email=None,
        )

    def _parse_contacts_table_address_cell(
        self, td: HtmlElement
    ) -> ContactAddress:
        """Parse the address TD of the table layout into a ContactAddress object.

        :param td: The td element to parse the address from.
        :return: A ContactAddress object.
        """

        raw_lines = []
        self._append_clean_text(td.text, raw_lines)
        for child in td:
            self._append_clean_text(child.text, raw_lines)
            self._append_clean_text(child.tail, raw_lines)

        lines = [
            ln for ln in (self._clean_whitespace(ln) for ln in raw_lines) if ln
        ]

        # Extract email line
        email = None
        for idx, ln in list(enumerate(lines)):
            if m := self.EMAIL_RE.search(ln):
                email = m.group(0)
                del lines[idx]
                break

        title, start_add_idx = self._parse_address_title(lines)
        partial_address = self._parse_address_lines(lines, start_add_idx)

        return ContactAddress(
            title=title,
            address_lines=partial_address.address_lines,
            city=partial_address.city,
            state=partial_address.state,
            zip_code=partial_address.zip_code,
            email=email,
        )

    def _section_by_heading(self, heading_text: str) -> Optional[HtmlElement]:
        """Return the "card-body" node for a given Contacts
        section heading.

        :param heading_text: The header text like "Attorneys for Petitioners".
        :return: The matching "card-body" node, or None if not found.
        """
        if self.tree is None:
            return None

        xpath_query = (
            "//div[@id='Contacts']//div[contains(@class,'card')]"
            f"[.//div[contains(@class,'card-heading') and normalize-space()='{heading_text}']]"
            "//div[contains(@class,'card-body')]"
        )

        return next(iter(self.tree.xpath(xpath_query)), None)

    def _build_attorney_from_contact(
        self, name_div: HtmlElement, data_div: Optional[HtmlElement]
    ) -> dict[str, Any]:
        """Build an attorney dict from ContactName/ContactData2 pair.

        :param name_div: ContactName div.
        :param data_div: ContactData2 div.
        :return: Attorney dict matching the JSON parser shape.
        """
        name_text = self._clean_whitespace(
            self._parse_description_html(name_div)
        )
        is_counsel_of_record = bool(
            name_div.xpath(
                ".//span[contains(@class,'ContactTitle')][contains(., 'Counsel of Record')]"
            )
        )

        title = address = city = state = zip_code = email = phone = None
        if data_div is not None:
            email = self._clean_whitespace(
                self._join_nodes_content(
                    data_div.xpath(".//span[contains(@class,'emailaddress')]")
                )
            )
            phone = self._clean_whitespace(
                self._join_nodes_content(
                    data_div.xpath(".//span[contains(@class,'phonenumber')]")
                )
            )
            phone = phone.replace("Ph:", "").strip() if phone else None

            address_span = data_div.xpath(
                ".//span[contains(@class,'address1')]"
            )
            if address_span:
                contact_data = self._parse_address_block(address_span[0])
                title = contact_data.title
                address_lines = contact_data.address_lines
                city = contact_data.city
                state = contact_data.state
                zip_code = contact_data.zip_code
                address = ", ".join(address_lines) if address_lines else None

        return {
            "name": name_text or None,
            "is_counsel_of_record": is_counsel_of_record,
            "title": title,
            "phone": phone,
            "address": address,
            "city": city,
            "state": state,
            "zip": zip_code,
            "email": email or None,
        }

    def _parse_address_block(self, node: HtmlElement) -> ContactAddress:
        """Parse the address block into title, address lines, city, state, zip.

        :param node: The HTML element (usually span.address1) to split by <br>.
        :return: A ContactAddress object.
        """

        raw_lines = []
        self._append_clean_text(node.text, raw_lines)
        for child in node:
            self._append_clean_text(child.text, raw_lines)
            self._append_clean_text(child.tail, raw_lines)

        if len(raw_lines) >= 3:
            # If more than 3 lines. Title is the first one.
            title = raw_lines[0]
            address_lines = raw_lines[1:]
        else:
            # 0, 1, or 2 lines, there is no title
            title = None
            address_lines = raw_lines[:]

        # Extract city/state/zip from the last address line, if present
        city = state = zip_code = None
        if address_lines:
            last = address_lines[-1]
            m = re.search(r"([^,]+),\s*([A-Z]{2})\s+(\d{5}(?:-\d{4})?)$", last)
            if m:
                city, state, zip_code = (
                    m.group(1).strip(),
                    m.group(2),
                    m.group(3),
                )
                address_lines = address_lines[:-1]

        return ContactAddress(
            title=title,
            address_lines=address_lines,
            city=city,
            state=state,
            zip_code=zip_code,
            email=None,
        )

    def _td_value_by_label(
        self, label: str, allow_indent: bool = False
    ) -> Optional[str]:
        """Fetch the value from the docketinfo table whose first TD label matches.

        :param label: The label text in the first TD.
        :param allow_indent: True to allow labels with leading indentation.
        :return: Cleaned string value or None.
        """
        if self.tree is None:
            return None

        if allow_indent:
            # Parse labels like "&nbsp;&nbsp;&nbsp;Case Numbers:"
            target_label = (
                f"//table[@id='docketinfo']//tr[td[1]/span[normalize-space(translate(text(), '\u00a0',' '))='{label}']]"
                "/td[2]//span"
            )
        else:
            target_label = f"//table[@id='docketinfo']//tr[td[1]/span[normalize-space(text())='{label}']]/td[2]//span"

        return self._clean_whitespace(
            self._join_nodes_content(self.tree.xpath(target_label))
        )

    @staticmethod
    def _clean_whitespace(text: Optional[str]) -> str:
        """Collapse whitespace to single spaces.

        :param text: Raw text.
        :return: Normalized text to a single-line.
        """
        return "" if not text else re.sub(r"\s+", " ", text).strip()

    @staticmethod
    def _join_nodes_content(nodes: list[HtmlElement]) -> str:
        """Concatenate the text content of multiple HTML nodes.

        :param nodes: List of lxml HtmlElement objects.
        :return: Combined and cleaned text content.
        """
        return "".join(
            node.text_content() for node in nodes if node is not None
        ).strip()


def _main():
    """Parse a local HTML file and pretty-print normalized data.

    :return: None
    """
    import pprint
    import sys

    if len(sys.argv) != 2:
        print(
            "Usage: python -m juriscraper.scotus_docket_report_html <filepath>"
        )
        sys.exit(1)

    parser = SCOTUSDocketReportHTML()
    with open(sys.argv[1], encoding="utf-8") as f:
        parser._parse_text(f.read())
    pprint.pprint(parser.data, indent=2)


if __name__ == "__main__":
    _main()



from juriscraper.opinions.united_states.federal_appellate import ca11_p


class Site(ca11_p.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "http://media.ca11.uscourts.gov/opinions/unpub/logname.php"
        self.court_id = self.__module__
        self.back_scrape_iterable = list(range(20, 22000, 20))
        self.should_have_results = True
        self.status = "Unpublished"



"""
CourtID: cadc
Court Short Name: Court of Appeals of the District of Columbia
Author: mlissner
History:
    2014-07-31, mlissner: commited first version
    2024-12-31, grossir: Implemented new site
    2025-08-26, lmanzur: Updated added extract_from_text.
"""

import re

from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        # https://media.cadc.uscourts.gov/opinions/
        self.url = "https://media.cadc.uscourts.gov/opinions/bydate/recent"
        self.status = "Published"

    def _process_html(self):
        link_xpath = "a[contains(@href, '.pdf')]"
        for row in self.html.xpath(f"//div[div[div[div[{link_xpath}]]]]"):
            self.cases.append(
                {
                    "url": row.xpath(f".//{link_xpath}/@href")[0],
                    "docket": row.xpath(f".//{link_xpath}/text()")[0],
                    "name": row.xpath("div[2]/div/div/text()")[0],
                    "date": row.xpath(".//span/text()")[-1],
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """

        pattern = re.compile(
            r"""
            (?:
               (?:On\s+)?Appeals?\s+from\s+the\s+
              | (?:On\s+)?Petitions?\s+for\s+Review\s+of\s+(?:a\s+)?(?:Final\s+)?Action\s+of\s+the\s+
            )
            (?P<lower_court>[^.]+?)
            (?=\s*(?:\.|\(|Nos?\.|USC|D.C.|\n\s*\n))
            """,
            re.X | re.IGNORECASE,
        )

        lower_court = ""
        if match := pattern.search(scraped_text):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()

        if lower_court:
            return {
                "Docket": {
                    "appeal_from_str": titlecase(lower_court),
                }
            }

        return {}



"""Scraper for the Air Force Court of Criminal Appeals
CourtID: afcca
Court Short Name: Air Force Court of Criminal Appeals
Reviewer: mlr
History:
    15 Sep 2014: Created by Jon Andersen
"""

from datetime import date

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    base_url = "http://afcca.law.af.mil/content/opinions_date_{}.html"
    start_year = 2002

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.current_year = date.today().year
        self.make_backscrape_iterable(kwargs)
        self.url = self.base_url.format(self.current_year)
        self.disable_certificate_verification()

    def _process_html(self) -> None:
        for row in self.html.xpath(".//img[contains(@src, 'pdf.gif')]/../..")[
            :-1
        ]:
            status = row.xpath(".//td")[3].text_content()
            if status != "Published":
                status = "Unpublished"
            self.cases.append(
                {
                    "url": row.xpath(".//td/a/@href")[0],
                    "docket": row.xpath(".//td")[2].text_content(),
                    "name": row.xpath(".//td/a")[0].text_content(),
                    "date": row.xpath(".//td")[4].text_content(),
                    "status": status,
                }
            )

    def _download_backwards(self, year: int) -> None:
        """Build URL with year input and scrape

        :param year: year to scrape
        :return None
        """
        self.url = self.base_url.format(year)
        self.html = self._download()
        self._process_html()

    def make_backscrape_iterable(self, kwargs: dict) -> None:
        """Checks if backscrape start and end arguments have been passed
        by caller, and parses them accordingly

        :param kwargs: passed when initializing the scraper, may or
            may not contain backscrape controlling arguments
        :return None
        """
        start = kwargs.get("backscrape_start")
        end = kwargs.get("backscrape_end")

        start = int(start) if start else self.start_year
        end = int(end) + 1 if end else self.current_year

        self.back_scrape_iterable = range(start, end)



from datetime import datetime

from . import alaska


class Site(alaska.Site):
    # Clarence Kameroff v State of Alaska. Opinion Number 1488
    first_opinion_date = datetime(1996, 10, 11).date()
    is_coa = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



"""
Scraper for Indiana Supreme Court
CourtID: ind
Court Short Name: Ind.
Auth: Jon Andersen <janderse@gmail.com>
Reviewer: mlr
History:
    2014-09-03: Created by Jon Andersen
    2025-07-09: Updated bt luism add new fields for lower court details and judge names
"""

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    page_court_id = "9510"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_name = "Supreme Court"
        self.status = "Published"
        self.url = f"https://public.courts.in.gov/Decisions/api/Search?courtId={self.page_court_id}"
        self.should_have_results = True

    def _process_html(self):
        """Process the HTML to extract case details.

        :return None
        """
        for case in self.html:
            lower_court, lower_court_number = self.parse_court_info(case)

            disposition = case["decision"]
            if disposition == "Opinion Issued":
                disposition = ""

            judge = ", ".join(
                self.clean_judge_name(v["judge"])
                for v in case["opinion"]["votes"]
            )

            per_curiam = case.get("opinion").get("perCuriam")
            author = (
                (
                    case["opinionText"]
                    .replace("in an opinion by ", "")
                    .replace(".", "")
                )
                if not per_curiam
                else ""
            )

            self.cases.append(
                {
                    "name": case["style"],
                    "docket": case["caseNumber"],
                    "date": case["date"],
                    "url": f"https://public.courts.in.gov/Decisions/{case['opinionUrl']}",
                    "disposition": disposition,
                    "lower_court": lower_court,
                    "lower_court_number": lower_court_number,
                    "judge": judge,
                    "author": author,
                    "per_curiam": case.get("opinion").get("perCuriam"),
                }
            )

    @staticmethod
    def clean_judge_name(name: str) -> str:
        """Cleans and formats a judge's name string.

        :param name: The name of the judge as a string.
        :return: A cleaned and formatted version of the judge's name.
        """

        parts = name.split(", ")
        if len(parts) < 2:
            return parts[0]

        result = parts[0]
        if "SR" in parts[1]:
            result += " SR"
        elif len(parts) > 2:
            result += f" {parts[2]}"

        return result

    def parse_court_info(self, case):
        """Extract the lower court name and number from case data

        :param case: case row json to parse
        :return: lower court data
        """
        excluded = {self.court_name}
        if self.court_name == "Court of Appeals":
            excluded.add("Supreme Court")

        # Filter out excluded courts
        courts = [
            c for c in case.get("courts", []) if c.get("name") not in excluded
        ]
        if not courts:
            return "", ""

        # Pick the court with the largest ID
        target = max(courts, key=lambda c: c.get("id", 0))
        name = target.get("name", "")
        number = target.get("number", "")

        # Normalize specific court names
        normalization = {
            "Court of Appeals": "Indiana Court of Appeals",
            "Tax Court": "Indiana Tax Court",
        }
        return normalization.get(name, name), number



# Scraper for California's Fifth District Court of Appeal
# CourtID: calctapp_5th
# Court Short Name: Cal. Ct. App.

from juriscraper.opinions.united_states.state import calctapp_1st


class Site(calctapp_1st.Site):
    court_code = "F"
    division = "5th App. Dist."



import re
from datetime import date, datetime

from dateutil.rrule import MONTHLY, rrule

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    lower_court_regexes = [
        re.compile(r"(?P<lower_court>U\.S\. District Court.+)"),
        re.compile(r"(?P<lower_court>Board of Immigration Appeals)"),
        re.compile(r"(?P<lower_court>U\.S\. Bankruptcy Court.+)"),
        re.compile(r"(?P<lower_court>Securities (&|and) Exchange Commission)"),
    ]
    first_opinion_date = datetime(1995, 11, 1)
    days_interval = 365 * 35  # we just need the start and end input dates
    base_url = "https://media.ca8.uscourts.gov/cgi-bin/opnByMM.pl?theMM=%02d&theYY=%s&A1=Get+Opinions"
    is_bap_scraper = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__

        today = date.today()
        self.url = self.base_url % (today.month, today.year)

        self.request["verify"] = False
        self.make_backscrape_iterable(kwargs)

    def _process_html(self):
        for link in self.html.xpath('//a[contains(@href, "opndir")]'):
            url = link.get("href")
            text = link.xpath("following-sibling::text()")[0].strip()

            lower_court, is_bap_opinion = self.get_lower_court(text)
            if is_bap_opinion:
                if not self.is_bap_scraper:
                    logger.info("Skipping `bap8` opinion %s", text)
                    continue
            elif self.is_bap_scraper:
                logger.info("Skipping `ca8` opinion %s", text)
                continue

            first_line = text.split("\n")[0]
            date_filed, case_name = first_line.strip().split(" ", 1)
            docket_number = ", ".join(re.findall(r"\d{2}-\d{4}", text))

            doc_name = link.text_content().split(".")[0].lower()
            if "p" in doc_name:
                status = "Published"
            elif "u" in doc_name:
                status = "Unpublished"
            else:
                status = "Unknown"

            self.cases.append(
                {
                    "date": date_filed,
                    "docket": docket_number,
                    "url": url,
                    "status": status,
                    "name": case_name.strip(),
                    "lower_court": lower_court,
                }
            )

    def get_lower_court(self, text: str) -> tuple[str, bool]:
        """Parse the lower court and identify if it is a bankruptcy court

        :param text: the raw lower court text
        :return: a tuple (lower_court, true if the lower court is a bankrupcy
            court)
        """
        for pattern in self.lower_court_regexes:
            if match := pattern.search(text):
                court = match.group("lower_court").strip()
                return court, "Bankruptcy Court" in court

        return "", False

    def make_backscrape_iterable(self, kwargs):
        """Parse the input dates, and create a tick for each month"""
        super().make_backscrape_iterable(kwargs)
        self.back_scrape_iterable = [
            i.date()
            for i in rrule(
                MONTHLY,
                dtstart=self.back_scrape_iterable[0][0],
                until=self.back_scrape_iterable[0][1],
            )
        ]

    def _download_backwards(self, d: date) -> None:
        """
        If an updated backscraper is needed in the future, this court
        updates the HTML with new values with a 4/5 months lag.
        Among the new values, I have seen the following fields we collect:
        per_curiam, judges and summaries are available
        """
        self.url = (
            "http://media.ca8.uscourts.gov/cgi-bin/opnByMM.pl?theMM=%02d&theYY=%s&A1=Get+Opinions"
            % (d.month, d.year)
        )
        logger.info("Backscraping %s-%s", d.year, d.month)
        self.html = self._download()
        self._process_html()



# Scraper for Texas Supreme Court
# CourtID: tex
# Court Short Name: TX
# Court Contacts:
#  - http://www.txcourts.gov/contact-us/
#  - Blake Hawthorne <Blake.Hawthorne@txcourts.gov>
#  - Eddie Murillo <Eddie.Murillo@txcourts.gov>
#  - Judicial Info <JudInfo@txcourts.gov>
# Author: Andrei Chelaru
# Reviewer: mlr
# History:
#  - 2014-07-10: Created by Andrei Chelaru
#  - 2014-11-07: Updated by mlr to account for new website.
#  - 2014-12-09: Updated by mlr to make the date range wider and more thorough.
#  - 2015-08-19: Updated by Andrei Chelaru to add backwards scraping support.
#  - 2015-08-27: Updated by Andrei Chelaru to add explicit waits
#  - 2021-12-28: Updated by flooie to remove selenium.
#  - 2024-02-21; Updated by grossir: handle dynamic backscrapes
#  - 2025-05-30; Updated by lmanzur: get opinions from the orders on causes page

import re
from datetime import date
from datetime import datetime as dt

from lxml import etree

from juriscraper.lib.string_utils import titlecase
from juriscraper.lib.type_utils import OpinionType
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    base_url = "https://www.txcourts.gov/supreme/orders-opinions/{}/"
    link_xp = '//*[@id="MainContent"]/div/div/div/ul/li/a/@href'
    date_xp = '//*[@id="MainContent"]/div/div[1]/div/text()'
    judge_xp = r"(?:Chief\s)?Justice\s([A-Z][a-zA-Z]+)"
    days_interval = 365

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.first_opinion_date = date(2014, 10, 3)
        self.current_year = date.today().year
        self.url = self.base_url.format(self.current_year)
        self.status = "Published"
        self.make_backscrape_iterable(kwargs)
        self.is_backscrape = False

    def _download(self, request_dict=None):
        """Downloads the HTML content for the current opinion page.

        :param request_dict (dict, optional): Additional request parameters.
        :return The downloaded HTML content.
        """
        if request_dict is None:
            request_dict = {}

        if not self.is_backscrape and not self.test_mode_enabled():
            self.html = super()._download(request_dict)
            self.url = self.html.xpath(self.link_xp)[-1]
        self.html = super()._download(request_dict)

        return self.html

    def _process_html(self) -> None:
        """Parses the HTML content

        :return None
        """
        date = self.html.xpath(self.date_xp)[0].strip()
        links = self.html.xpath('//a[contains(@href, ".pdf")]')
        for link in links:
            if link.getparent() is None or link.getparent().get(
                "class"
            ) not in ["a79", "a70"]:
                continue
            precedingTRs = link.xpath(
                'ancestor::tr/preceding-sibling::tr[td[@class="a50cl"]]'
            ) or link.xpath(
                'ancestor::tr/preceding-sibling::tr[td[@class="a54cl"]]'
            )

            docket, title, *_ = [
                tr_text
                for tr_text in precedingTRs[-1].xpath(".//text()")
                if tr_text.strip()
            ]

            disposition = (
                link.getparent().xpath(".//text()")[0]
                if link.getparent().get("class") == "a70"
                else ""
            )

            judge_str = (
                "".join(
                    [
                        x
                        for x in [
                            link.getprevious().tail,
                            link.text,
                            link.tail,
                        ]
                        if x
                    ]
                )
                if disposition
                else ""
            )

            judges = re.findall(self.judge_xp, judge_str)
            if judges:
                author = judges[0]
                per_curiam = False
            else:
                author = ""
                per_curiam = True

            lower_court, lower_court_number, lower_court_id = (
                self.parse_lower_court_info(title)
            )

            title_regex = r"^(?P<name>.*?)(?=; from|(?=\([^)]*\)$))"
            title_match = re.search(title_regex, title, flags=re.MULTILINE)

            self.cases.append(
                {
                    "name": titlecase(
                        title_match.group("name") if title_match else title
                    ),
                    "disposition": disposition,
                    "url": link.get("href"),
                    "docket": docket,
                    "date": date,
                    "type": self.extract_type(link),
                    "per_curiam": per_curiam,
                    "judge": ", ".join(judges),
                    "author": author,
                    "lower_court": lower_court,
                    "lower_court_number": lower_court_number,
                    "lower_court_id": lower_court_id,
                }
            )

    @staticmethod
    def parse_lower_court_info(title: str) -> tuple[str, str, str]:
        """Parses lower court information from the title string

        :param title: a string with lower court information
        :return: values for lower_court, lower_court_number, lower_court_id
        """

        # Format when appeal comes from texapp districts. Example:
        # ' from Harris County; 1st Court of Appeals District (01-22-00182-CV, 699 SW3d 20, 03-23-23)'
        texapp_regex = r" from (?P<lower_court>.*)\s*\("

        # Format when appeal comes from other possible courts. Examples:
        #  "(U.S. Fifth Circuit 23-10804)"
        #  "(U.S. 5th Circuit 19-51012)"
        # "(BODA Cause No. 67623)"
        other_courts_regex = r"\((?P<lower_court>(BODA|U\.S\. (Fif|5)th Circuit))\s(?P<lower_number>(Cause No. )?[\d-]+)\)$"

        if match := re.search(texapp_regex, title):
            lower_court = match.group("lower_court")
            lower_court_number = title[match.end() :].split(",")[0]
            return lower_court, lower_court_number, "texapp"

        elif match := re.search(other_courts_regex, title):
            lower_court = match.group("lower_court")
            lower_court_number = match.group("lower_number")

            if lower_court == "BODA":
                lower_court = "Board of Disciplinary Appeals"
                lower_court_id = "txboda"
            else:
                # if this is not a BODA match, then it can only be a
                # Fifth Circuit match. Update this if the regex above changes
                lower_court_id = "ca5"

            return lower_court, lower_court_number, lower_court_id

        return "", "", ""

    @staticmethod
    def extract_type(link: etree.Element) -> str:
        """Determines the opinion type

        :param link (etree.Element) The anchor element containing the PDF link.
        :return str The opinion type as a string.
        """
        text = link.text.lower()
        url = link.get("href")
        if "per curiam" in text or url.endswith("pc.pdf"):
            op_type = OpinionType.UNANIMOUS
        elif "in part" in text or url.endswith("cd.pdf"):
            op_type = OpinionType.CONCURRING_IN_PART_AND_DISSENTING_IN_PART
        elif "dissenting" in text or url.endswith("d.pdf"):
            op_type = OpinionType.DISSENT
        elif "concurring" in text or url.endswith("c.pdf"):
            op_type = OpinionType.CONCURRENCE
        else:
            op_type = OpinionType.MAJORITY
        return str(op_type.value)

    def make_backscrape_iterable(self, kwargs) -> None:
        """Checks if backscrape start and end arguments have been passed

        Texas opinions page returns all opinions for a year (pagination is not needed),
        so we must filter out opinions not in the date range we are looking for

        :return None
        """
        super().make_backscrape_iterable(kwargs)

        # use the parsed values to compute the actual iterable
        start = self.back_scrape_iterable[0][0]
        end = self.back_scrape_iterable[-1][-1]

        dates = []
        for year in list(range(start.year, end.year + 1)):
            dates.append((year, start, end))
        self.back_scrape_iterable = dates

    def _download_backwards(self, analysis_window: tuple) -> None:
        """Downloads and processes opinions for a given year within a specified date range.

        :param analysis_window (tuple): A tuple containing the year (int), start date (date), and end date (date).
        :return None
        """
        self.is_backscrape = True
        year, start, end = analysis_window
        self.url = self.base_url.format(year)
        self._download()

        for path in self.html.xpath(self.link_xp):
            if "historical" in path:
                # in 2014 they have an extra link for pre-2014 ops
                continue
            date_str = path.strip("/").split("/")[-1]
            date_obj = dt.strptime(date_str, "%B-%d-%Y").date()

            if not (start <= date_obj <= end):
                continue

            self.url = path
            self._download()
            self._process_html()



# Scraper for Oklahoma Court of Criminal Appeals
# CourtID: oklacrimapp
# Court Short Name: OK
# Author: Andrei Chelaru
# Reviewer: mlr
# Date: 2014-07-05


from juriscraper.opinions.united_states.state import okla


class Site(okla.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.oscn.net/decisions/ok-cr/90"



# Scraper for Oklahoma Attorney General Opinions
# CourtID: oklaag
# Court Short Name: OK
# Author: Andrei Chelaru
# Reviewer: mlr
# Date: 2014-07-05

from datetime import datetime

from juriscraper.opinions.united_states.state import okla


class Site(okla.Site):
    # Inherit cleanup_content from Okla
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        year = datetime.today().year
        self.url = f"https://www.oscn.net/applications/oscn/Index.asp?ftdb=STOKAG&year={year}&level=1"
        self.status = "Published"
        self.expected_content_types = ["text/html"]

    def _process_html(self):
        for row in self.html.xpath("//div/p['@class=document']")[::-1]:
            if "OK" not in row.text_content() or "EMAIL" in row.text_content():
                continue
            citation, date, name = row.text_content().split(",", 2)
            self.cases.append(
                {
                    "date": date,
                    "name": name,
                    "docket": "",
                    "url": row.xpath(".//a")[0].get("href"),
                    "citation": citation,
                }
            )



import re

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://legacy.utcourts.gov/opinions/supopin/"
        self.court_id = self.__module__
        self.status = "Published"
        self.should_have_results = True

    def _process_html(self):
        for row in self.html.xpath(
            "//div[@id='content']//p[a[contains(@href, '.pdf')]]"
        ):
            texts = row.xpath("text()")
            if any("superseded" in t.lower() for t in texts):
                # Superseding opinions are published in their own rows
                logger.info("Skipping row with superseded opinion %s", texts)
                continue

            # pick longest text; if not, HTML comments may cause wrong indexing
            text = sorted(texts)[-1]
            neutral_cite_match = re.search(r"\d{4} UT( App)? \d{1,}", text)
            citation = neutral_cite_match.group(0)

            filed_index = text.find("Filed")
            docket = text[:filed_index].strip(", ")
            date_filed = text[
                filed_index + 5 : neutral_cite_match.start()
            ].strip(" ,")

            self.cases.append(
                {
                    "url": row.xpath("a")[0].get("href"),
                    "name": row.xpath("a")[0].text_content(),
                    "date": date_filed,
                    "citation": citation,
                    "docket": docket,
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """
        pattern = re.compile(
            r"""
            (On\s+Direct\s+Appeal
                (?:\s+and\s+Petition\s+for\s+Extraordinary\s+Relief)?
            |On\s+Appeal\s+of\s+Interlocutory\s+Order
            |On\s+Certification\s+from\s+the\s+Court\s+of\s+Appeals
            |On\s+Petition\s+for\s+Review\s+ of\s+Agency\s+Decision
            |On\s+Certiorari\s+to\s+the\s+Utah\s+Court\s+of\s+Appeals)

            \s*\n+
            (?P<lower_court>([^\S\r\n]*\S.*(?:\n|$))+?)
            [^\S\r\n]*the\s+Honorable\s+(?P<lower_court_judge>.*?)(\s+No\.)\s*(?P<lower_court_number>\S+)
            """,
            re.MULTILINE | re.VERBOSE | re.IGNORECASE,
        )

        if match := pattern.search(scraped_text):
            lower_court_str = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()
            lower_court_judge = match.group("lower_court_judge").strip()
            lower_court_number = match.group("lower_court_number").strip()

            return {
                "Docket": {
                    "appeal_from_str": lower_court_str,
                },
                "OriginatingCourtInformation": {
                    "assigned_to_str": lower_court_judge,
                    "docket_number": lower_court_number,
                },
            }

        return {}



from juriscraper.opinions.united_states.state import utah
from juriscraper.OpinionSite import OpinionSite


class Site(utah.Site):
    extract_from_text = OpinionSite.extract_from_text

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://legacy.utcourts.gov/opinions/appopin/"
        self.court_id = self.__module__



from .scotus_docket_report import SCOTUSDocketReport
from .scotus_docket_report_htm import SCOTUSDocketReportHTM
from .scotus_docket_report_html import SCOTUSDocketReportHTML

__all__ = [SCOTUSDocketReport, SCOTUSDocketReportHTML, SCOTUSDocketReportHTM]



import re
from datetime import date, datetime

from dateutil.relativedelta import relativedelta

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    oldest_opinion = "2002-03-20"
    court_name = "United States Court of Appeals for the Fourth Circuit"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://www.govinfo.gov/wssearch/search"
        self.court_id = self.__module__
        self.method = "POST"
        self.expected_content_types = ["application/pdf"]

        self.start = (date.today() - relativedelta(months=1)).strftime(
            "%Y-%m-%d"
        )
        self.end = date.today().strftime("%Y-%m-%d")
        self.date_range = f"{self.start},{self.end}"
        self.parameters = {}
        self.update_parameters()
        self.make_backscrape_iterable(kwargs)

    def make_backscrape_iterable(self, kwargs: dict[str, str]) -> None:
        """Make back scrape iterable

        :param kwargs: the back scraping params
        :return: None
        """
        start_str = kwargs.get("backscrape_start", self.oldest_opinion)
        end_str = kwargs.get("backscrape_end", self.end)

        start_date = datetime.strptime(start_str, "%Y-%m-%d").date()
        end_date = datetime.strptime(end_str, "%Y-%m-%d").date()

        date_ranges: list[str] = []
        current = start_date

        while current <= end_date:
            week_end = min(current + relativedelta(days=7), end_date)
            date_ranges.append(f"{current:%Y-%m-%d},{week_end:%Y-%m-%d}")
            current = week_end + relativedelta(days=1)

        self.back_scrape_iterable = date_ranges

    def _process_html(self) -> None:
        """Process CA4 Opinions

        :return: None
        """
        for row in self.html["resultSet"]:
            line2 = row.get("line2")
            if (
                "OPINION" not in line2
                or "day, " not in line2
                or "OPINION ATTACHMENT" in line2
            ):
                logger.debug("Skipping %s", line2)
                continue

            teaser = row.get("fieldMap").get("teaser")
            metadata = self.extract_metadata(teaser)
            is_published = " PUBLISHED" in line2
            case = {
                "docket": row["line1"].split()[0],
                "name": row.get("fieldMap").get("title"),
                "url": row.get("fieldMap").get("url"),
                "date": row["line2"].split("day, ")[1].strip("."),
                "status": "Published" if is_published else "Unpublished",
                "per_curiam": "PER CURIAM" in line2,
            }
            case.update(metadata)
            self.cases.append(case)

    def extract_metadata(self, teaser: str) -> dict[str, str]:
        """Extract out various metadata fields

        Method leaves in a few regex patterns that we do not currently support
        but could be added to our ingestion pipeline.

        :param teaser: The teaser line provided in some opinions
        :return: Extracted data
        """
        if not teaser:
            return {"lower_court": "", "lower_court_number": ""}
        patterns = {
            "lower_court": re.compile(
                r"""
                (?:
                   Appeal(?:s)?\s+from\s+the\s+
                 | On\s+Petition\s+for\s+Review\s+of\s+an\s+Order\s+of\s+the\s+
                )
                (?P<lower_court>[^.,]+?)
                (?=[\.,])
                """,
                re.X,
            ),
            "lower_court_number": re.compile(
                r"""
                \(\s*
                (?P<lower_court_number>\d+:\d{2}-[a-z]{2}-\d+(?:-[\w-]+)*)
                \s*\)
                """,
                re.IGNORECASE | re.VERBOSE,
            ),
            # "date_argued": re.compile(
            #     r"Argued:\s*(?P<date_argued>[A-Z][a-z]+ \d{1,2}, \d{4})"),
            # "date_decided": re.compile(
            #     r"Decided:\s*(?P<date_decided>[A-Z][a-z]+ \d{1,2}, \d{4})"),
            # "amended_date": re.compile(
            #     r"Amended:\s*(?P<amended_date>[A-Z][a-z]+ \d{1,2}, \d{4})"),
            # "panel_str": re.compile(r"Before\s+(?P<panel_str>[^.]+)"),
        }
        metadata: dict[str, str] = {}
        for name, pat in patterns.items():
            m = pat.search(teaser)
            match = m.group(name) if m else ""
            if (
                match.strip()
                == "United States District Court for the Southern District of West"
            ):
                match = "United States District Court for the Southern District of West Virginia"
            metadata[name] = match
        return metadata

    def update_parameters(self):
        """Update the date range parameter"""

        self.request["parameters"]["json"] = {
            "historical": True,
            "offset": 0,
            "query": f"collection:(USCOURTS) AND publishdate:range({self.date_range})  AND  courtname:(United States Court of Appeals for the Fourth Circuit)",
            "pageSize": 100,
            "sortBy": "2",  # 2 -> newest to oldest
        }

    def _download_backwards(self, date_range) -> None:
        """Download backward

        :param date_range: the date range as a string
        :return: None
        """
        self.date_range = date_range
        self.update_parameters()
        self.html = self._download()
        self._process_html()



"""Scraper for D.C. Circuit of Appeals
CourtID: cadc
Court Short Name: cadc
Author: Andrei Chelaru
Reviewer: mlr
Date created: 18 July 2014
Updated: 2024-10-10
"""

from datetime import date, datetime
from urllib.parse import urljoin

from juriscraper.AbstractSite import logger
from juriscraper.lib.date_utils import unique_year_month
from juriscraper.OralArgumentSiteLinear import OralArgumentSiteLinear


class Site(OralArgumentSiteLinear):
    days_interval = 28  # ensure monthly backscraper ticks
    first_opinion_date = datetime(2007, 9, 10)
    base_url = "https://media.cadc.uscourts.gov/recordings/bydate/{}"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        today = date.today()
        self.url = self.base_url.format(today.strftime("%Y/%-m"))
        self.request["verify"] = False
        self.make_backscrape_iterable(kwargs)

    def _process_html(self):
        anchor_xpath = "a[contains(@href, '/recordings/docs/') and contains(@href, '.mp3')]"
        for row in self.html.xpath(f"//div[div[div[div[{anchor_xpath}]]]]"):
            ahref = row.xpath(f".//{anchor_xpath}")
            url = ahref[0].xpath("@href")[0]
            docket = ahref[0].xpath("text()")[0]

            if "mp3" not in url:
                logger.info("Not a mp3 URL? %s", url)
                continue

            cells = row.xpath("div[2]/div")

            # From: "Judges: Tatel, Griffith, Sentelle"
            # To: "Tatel; Griffith; Sentelle"
            judges = (
                cells[1].text_content().split(": ", 1)[-1].replace(",", ";")
            )
            self.cases.append(
                {
                    "url": urljoin(self.base_url, url),
                    "docket": docket,
                    "name": cells[0].text_content(),
                    "judge": judges,
                    "attorney": cells[2].text_content(),
                    "date": cells[3].text_content(),
                }
            )

    def _download_backwards(self, target_date: date) -> None:
        """Download historical data

        Note that this URL will work:
        "https://media.cadc.uscourts.gov/recordings/bydate/2007/9"
        but this won't
        "https://media.cadc.uscourts.gov/recordings/bydate/2007/09"

        That's why the '%-m' formatter is needed
        """
        self.url = self.base_url.format(target_date.strftime("%Y/%-m"))
        logger.info("Backscraping URL '%s'", self.url)
        self.html = self._download()
        self._process_html()

    def make_backscrape_iterable(self, kwargs: dict) -> None:
        """Use base function to generate a range, then pick
        unique year-month combinations to build the backscrape
        URLS
        """
        super().make_backscrape_iterable(kwargs)
        self.back_scrape_iterable = unique_year_month(
            self.back_scrape_iterable
        )



from datetime import date

from juriscraper.opinions.united_states.state import sc


class Site(sc.Site):
    court = "court-of-appeals"
    opinion_status = "unpublished-opinions"
    first_opinion_date = date(2003, 1, 1)



import re
from datetime import date, datetime, timedelta

from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    lower_court_regex = re.compile(r"FROM THE (?P<lower_court>.+)")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "http://www.courts.state.va.us/scndex.htm"
        self.status = "Published"
        self.should_have_results = True

    def _process_html(self) -> None:
        """Parses the HTML content to extract case information.

        :return: None
        """
        if self.test_mode_enabled():
            today = datetime.strptime("11/20/2023", "%m/%d/%Y").date()
        else:
            today = date.today()

        for row in self.html.xpath("//p"):
            links = row.xpath(".//a")
            if len(links) != 2:
                continue
            name = row.xpath(".//b/text()")[0].strip()

            text = row.text_content().split(name)[1].strip()
            split_test = text.split("\n")

            date_str = split_test[0].strip()
            summary = split_test[1].strip()

            if "Revised" in date_str:
                date_str = date_str.split("Revised")[1].strip().strip(")")
            date_object = datetime.strptime(date_str, "%m/%d/%Y").date()
            if date_object < today - timedelta(days=60):
                # Stop after two months
                break

            self.cases.append(
                {
                    "name": row.xpath(".//b/text()")[0].strip(),
                    "docket": links[1].text.strip(" '\""),
                    "url": links[1].get("href"),
                    "summary": summary,
                    "date": date_str,
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """
        if match := self.lower_court_regex.search(scraped_text[:1000]):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()
            return {
                "Docket": {
                    "appeal_from_str": titlecase(lower_court),
                }
            }

        return {}



from juriscraper.lib.string_utils import convert_date_string
from juriscraper.OpinionSite import OpinionSite


class OpinionSiteLinear(OpinionSite):
    """This class can be used for any site that needs to be scraped linearly,
    instead of, for example, with separate html path parsing getters. Sometimes
    it is just easier and less repetitive to scrape a site this way, in which
    case you can simply extend this class and implement _process_html().
    """

    # This class tries to simplify OpinionSiteLinear. Instead of using full
    # attribute names, we use shorthands. The keys must be properly named
    # for the getters to work, so this `valid_keys` will be used in an
    # extended check_sanity method
    valid_keys = {
        "name",
        "url",
        "date",
        "date_filed_is_approximate",
        "status",
        "docket",
        "judge",
        "citation",
        "parallel_citation",
        "summary",
        "lower_court",
        "child_court",
        "adversary_number",
        "division",
        "disposition",
        "cause",
        "docket_attachment_number",
        "docket_document_number",
        "nature_of_suit",
        "lower_court_id",
        "lower_court_number",
        "lower_court_judge",
        "author",
        "per_curiam",
        "type",
        "joined_by",
        "other_date",
        "attorney",
        "headnote",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cases = []
        self.status = None

    def _process_html(self):
        raise Exception(
            "Must implement _process_html() on OpinionSiteLinear child"
        )

    def _get_case_names(self):
        return [case["name"] for case in self.cases]

    def _get_download_urls(self):
        return [case["url"] for case in self.cases]

    def _get_case_dates(self):
        return [convert_date_string(case["date"]) for case in self.cases]

    def _get_date_filed_is_approximate(self):
        return [
            case.get("date_filed_is_approximate", False) for case in self.cases
        ]

    def _get_precedential_statuses(self):
        # first try to use status values set in cases dictionary
        try:
            return [case["status"] for case in self.cases]
        except AttributeError:
            pass
        except KeyError:
            pass
        # we fall back on using singular status defined in init,
        # which is all you need to do if all cases on the page
        # have the same status
        if not self.status:
            raise Exception(
                "Must define self.status in __init__ on OpinionSiteLinear child"
            )
        return [self.status] * len(self.cases)

    def _get_docket_numbers(self):
        return [case["docket"] for case in self.cases]

    # optional getters below

    def _get_optional_field_by_id(self, id):
        if self.cases and id in self.cases[0]:
            return [case[id] for case in self.cases]

    def _get_judges(self):
        return self._get_optional_field_by_id("judge")

    def _get_citations(self):
        return self._get_optional_field_by_id("citation")

    def _get_parallel_citations(self):
        return self._get_optional_field_by_id("parallel_citation")

    def _get_summaries(self):
        return self._get_optional_field_by_id("summary")

    def _get_child_courts(self):
        return self._get_optional_field_by_id("child_court")

    def _get_lower_courts(self):
        return self._get_optional_field_by_id("lower_court")

    def _get_lower_court_ids(self):
        return self._get_optional_field_by_id("lower_court_id")

    def _get_lower_court_judges(self):
        return self._get_optional_field_by_id("lower_court_judge")

    def _get_lower_court_numbers(self):
        return self._get_optional_field_by_id("lower_court_number")

    def _get_nature_of_suit(self):
        return self._get_optional_field_by_id("nature_of_suit")

    def _get_docket_document_numbers(self):
        return self._get_optional_field_by_id("docket_document_number")

    def _get_docket_attachment_numbers(self):
        return self._get_optional_field_by_id("docket_attachment_number")

    def _get_causes(self):
        return self._get_optional_field_by_id("cause")

    def _get_dispositions(self):
        return self._get_optional_field_by_id("disposition")

    def _get_divisions(self):
        return self._get_optional_field_by_id("division")

    def _get_adversary_numbers(self):
        return self._get_optional_field_by_id("adversary_number")

    def _get_authors(self):
        return self._get_optional_field_by_id("author")

    def _get_per_curiam(self):
        return self._get_optional_field_by_id("per_curiam")

    def _get_joined_by(self):
        return self._get_optional_field_by_id("joined_by")

    def _get_types(self):
        return self._get_optional_field_by_id("type")

    def _get_other_dates(self):
        """Goes into OpinionCluster.other_dates, type: string"""
        return self._get_optional_field_by_id("other_date")

    def _get_attorneys(self):
        """Goes into OpinionCluster.attorneys, type: string"""
        return self._get_optional_field_by_id("attorney")

    def _get_headnotes(self):
        """Goes into OpinionCluster.headnotes, type: string"""
        return self._get_optional_field_by_id("headnote")

    def _check_sanity(self):
        super()._check_sanity()
        # Check that all returned keys have the proper name to be used
        # in an Opinion / AbstractSite getter
        for case in self.cases:
            if isinstance(case, str):
                # happens with mich example files
                continue

            for key in case:
                if key not in self.valid_keys:
                    raise KeyError(
                        f"Invalid key '{key}' for case dictionary {self.__module__}"
                    )



"""
Scraper for Alabama Supreme Court
CourtID: ala
Court Short Name: Alabama
Author: William Palin
Court Contact:
History:
 - 2023-01-04: Created.
 - 2023-011-14: Alabama no longer uses page or use selenium.
"""

import re

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    court_str = "68f021c4-6a44-4735-9a76-5360b2e8af13"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = f"https://publicportal-api.alappeals.gov/courts/cms/publications?courtID={self.court_str}&page=0&size=25&sort=publicationDate%2Cdesc"
        self.should_have_results = True

    def _process_html(self):
        self.json = self.html

        # Processes only the first result to scrape the most recent data.
        item = self.json["_embedded"]["results"][0]

        date_filed = item["scheduledDate"][:10]
        for publicationItem in item["publicationItems"]:
            if not publicationItem.get("documents", []):
                continue

            url = f"https://publicportal-api.alappeals.gov/courts/{self.court_str}/cms/case/{publicationItem['caseInstanceUUID']}/docketentrydocuments/{publicationItem['documents'][0]['documentLinkUUID']}"
            docket = publicationItem["caseNumber"]
            name = publicationItem["title"]

            lower_court = ""
            lower_court_number = ""
            # Regex to match: (Appeal from <court>: <number>) or (Appeal from <court>: <number> and <number>)
            match = re.search(
                r"\(Appeal from (?P<lower_court>.+?): (?P<lower_court_number>.+?)\)",
                name,
            )
            if match:
                lower_court = match.group("lower_court").strip()
                lower_court_number = match.group("lower_court_number").strip()
                # Remove the parenthetical from the name
                name = name[: match.start()].rstrip()

            judge = publicationItem["groupName"]
            if judge == "On Rehearing":
                judge = ""

            per_curiam = False
            if "curiam" in judge.lower():
                judge = ""
                per_curiam = True

            self.cases.append(
                {
                    "date": date_filed,
                    "name": name,
                    "docket": docket,
                    "status": "Published",
                    "url": url,
                    "judge": judge,
                    "per_curiam": per_curiam,
                    "lower_court": lower_court,
                    "lower_court_number": lower_court_number,
                }
            )



# Scraper for Texas Supreme Court
# CourtID: tex
# Court Short Name: TX
# Court Contacts:
#  - http://www.txcourts.gov/contact-us/
#  - Blake Hawthorne <Blake.Hawthorne@txcourts.gov>
#  - Eddie Murillo <Eddie.Murillo@txcourts.gov>
#  - Judicial Info <JudInfo@txcourts.gov>
# Author: Andrei Chelaru
# Reviewer: mlr
# History:
#  - 2014-07-10: Created by Andrei Chelaru
#  - 2014-11-07: Updated by mlr to account for new website.
#  - 2014-12-09: Updated by mlr to make the date range wider and more thorough.
#  - 2015-08-19: Updated by Andrei Chelaru to add backwards scraping support.
#  - 2015-08-27: Updated by Andrei Chelaru to add explicit waits
#  - 2021-12-28: Updated by flooie to remove selenium.
#  - 2024-02-21; Updated by grossir: handle dynamic backscrapes

from datetime import date, datetime, timedelta
from typing import Optional

from juriscraper.AbstractSite import logger
from juriscraper.DeferringList import DeferringList
from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    param_date_format = "%-m/%-d/%Y"
    first_opinion_date = datetime(2002, 1, 24, 0, 0, 0)
    # Interval for default scrape and backscrape iterable generation
    days_interval = 15

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.checkbox = 0
        self.status = "Published"
        self.url = "https://search.txcourts.gov/CaseSearch.aspx?coa=cossup"
        self.make_backscrape_iterable(kwargs)
        self.next_page = None
        self.current_page = 1
        # Default scrape range if not doing a backscrape
        self.end_date = date.today()
        self.start_date = self.end_date - timedelta(days=self.days_interval)
        self.is_first_request = True
        self.seeds = []

    def _set_parameters(self) -> None:
        """Set ASPX post parameters

        This method - chooses the court and date parameters.
        ctl00$ContentPlaceHolder1$chkListCourts$[KEY] is what selects a ct

         0: Supreme
         1: Court of Criminal Appeals
         2: 1st District Court of Appeals
         3: 2nd District Court of Appeals
         ...etc

        :return: None
        """
        start_date_str = self.start_date.strftime(self.param_date_format)
        end_date_str = self.end_date.strftime(self.param_date_format)

        from_date_param = self.make_date_param(self.start_date, start_date_str)
        to_date_param = self.make_date_param(self.end_date, end_date_str)

        self.parameters = {}
        for hidden in self.html.xpath("//input[@type='hidden']"):
            value = hidden.xpath("@value")[0] if hidden.xpath("@value") else ""
            self.parameters[hidden.xpath("@name")[0]] = value

        self.parameters.update(
            {
                "ctl00$ContentPlaceHolder1$SearchType": "rbSearchByDocument",  # "Document Search" radio button
                "ctl00$ContentPlaceHolder1$dtDocumentFrom": str(
                    self.start_date
                ),
                "ctl00$ContentPlaceHolder1$dtDocumentFrom$dateInput": start_date_str,
                "ctl00$ContentPlaceHolder1$dtDocumentTo": str(self.end_date),
                "ctl00$ContentPlaceHolder1$dtDocumentTo$dateInput": end_date_str,
                "ctl00_ContentPlaceHolder1_dtDocumentFrom_dateInput_ClientState": from_date_param,
                "ctl00_ContentPlaceHolder1_dtDocumentTo_dateInput_ClientState": to_date_param,
                "ctl00$ContentPlaceHolder1$btnSearchText": "Search",
                "ctl00$ContentPlaceHolder1$chkListDocTypes$0": "on",  # "Opinion" checkbox
                f"ctl00$ContentPlaceHolder1$chkListCourts${self.checkbox}": "on",  # Court checkbox
                "ctl00$ContentPlaceHolder1$txtSearchText": "",
                "ctl00_ContentPlaceHolder1_dtDocumentFrom_ClientState": '{"minDateStr":"1900-01-01-00-00-00","maxDateStr":"2099-12-31-00-00-00"}',
                "ctl00_ContentPlaceHolder1_dtDocumentTo_ClientState": '{"minDateStr":"1900-01-01-00-00-00","maxDateStr":"2099-12-31-00-00-00"}',
                "ctl00$ContentPlaceHolder1$Stemming": "on",
                "ctl00$ContentPlaceHolder1$Fuzziness": "0",
            }
        )

        # "Next page" arrow button has a name like
        # "ctl00$ContentPlaceHolder1$grdDocuments$ctl00$ctl02$ctl00$ctl18"
        # Couldn't get pagination working when using the numbered page anchors,
        # whose id goes into __EVENTTARGET
        if self.next_page:
            self.parameters[self.next_page[0].xpath("@name")[0]] = ""

    def _process_html(self) -> None:
        """Process HTML and paginates if needed

        :return: None
        """
        if not self.test_mode_enabled():
            # Make our post request to get our data
            self.method = "POST"
            self._set_parameters()
            self.html = super()._download()

        rows_xpath = "//table[@class='rgMasterTable']/tbody/tr[not(@class='rgNoRecords')]"
        for row in self.html.xpath(rows_xpath):
            # In texas we also have to ping the case page to get the name
            # this is unfortunately part of the process.
            self.seeds.append(row.xpath(".//a")[2].get("href"))
            self.cases.append(
                {
                    "date": row.xpath("td[2]")[0].text_content(),
                    "docket": row.xpath("td[5]")[0].text_content().strip(),
                    "url": row.xpath(".//a")[1].get("href"),
                }
            )

        next_page_xpath = (
            "//input[@title='Next Page' and not(@onclick='return false;')]"
        )
        self.next_page = self.html.xpath(next_page_xpath)
        if self.next_page and not self.test_mode_enabled():
            self.current_page += 1
            logger.info(f"Paginating to page {self.current_page}")
            self._process_html()

    def _get_case_names(self) -> DeferringList:
        """Get case names using a deferring list."""

        def get_name(link: str) -> Optional[str]:
            """Abstract out the case name from the case page."""
            if self.test_mode_enabled():
                return "No case names fetched during tests."
            html = self._get_html_tree_by_url(link)
            try:
                plaintiff = html.xpath(
                    '//label[contains(text(), "Style:")]/parent::div/following-sibling::div/text()'
                )[0].strip()
                defendant = html.xpath(
                    '//label[contains(text(), "v.:")]/parent::div/following-sibling::div/text()'
                )[0].strip()

                # In many cases the court leaves off the plaintiff (The State of Texas).  But
                # in some of these cases the appellant is ex parte.  So we need to
                # add the state of texas in some cases but not others.
                if defendant:
                    return titlecase(f"{plaintiff} v. {defendant}")
                elif "WR-" not in link:
                    return titlecase(f"{plaintiff} v. The State of Texas")
                else:
                    return titlecase(f"{plaintiff}")
            except IndexError:
                logger.warning(f"No title or defendant found for {self.url}")
                return None

        return DeferringList(seed=self.seeds, fetcher=get_name)

    def _download_backwards(self, dates: tuple[date]) -> None:
        """Overrides present scraper start_date and end_date

        :param dates: (start_date, end_date) tuple
        :return None
        """
        start, end = dates
        self.start_date = (
            start.date() if not isinstance(start, date) else start
        )
        self.end_date = end.date() if not isinstance(end, date) else end
        logger.info(
            "Backscraping for range %s %s", self.start_date, self.end_date
        )

    @staticmethod
    def make_date_param(date_obj: date, date_str: str) -> str:
        """Make JSON encoded string with dates as expected by formdata

        :param date_obj: a date object
        :param date_str: string representation of the date in expected format

        :return: JSON encoded string
        """
        return (
            '{"enabled":true,"emptyMessage":"",'
            f'"validationText":"{date_obj}-00-00-00",'
            f'"valueAsString":"{date_obj}-00-00-00",'
            '"minDateStr":"1900-01-01-00-00-00",'
            '"maxDateStr":"2099-12-31-00-00-00",'
            f'"lastSetTextBoxValue":"{date_str}"'
            "}"
        )



import re
from collections import defaultdict
from typing import Any, Optional

from lxml import html
from lxml.html import HtmlElement

from juriscraper.lib.html_utils import strip_bad_html_tags_insecure
from juriscraper.lib.log_tools import make_default_logger
from juriscraper.lib.string_utils import clean_string, harmonize

from .scotus_docket_report_html import SCOTUSDocketReportHTML

logger = make_default_logger()


class SCOTUSDocketReportHTM(SCOTUSDocketReportHTML):
    """Parse SCOTUS docket HTM pages."""

    CONTACTS_TABLE_XPATH = (
        "//table[.//tr[td[1][contains(normalize-space(.), '~~Name')] "
        "and td[2][contains(normalize-space(.), '~~~~~~~Address')] "
        "and td[3][contains(normalize-space(.), '~~Phone')]]]"
    )
    HEADER_ROW_XPATH = (
        ".//tr[td[1][contains(normalize-space(.), '~~Name')] "
        "and td[2][contains(normalize-space(.), '~~~~~~~Address')] "
        "and td[3][contains(normalize-space(.), '~~Phone')]]"
    )

    def __init__(self, court_id: str = "scotus"):
        """Initialize the HTM report parser.

        :param court_id: The court ID.
        :return: None.
        """
        super().__init__(court_id=court_id)

    @property
    def metadata(self) -> dict[str, Any]:
        """Return normalized metadata extracted from the HTM docket.

        :return: A dict with the SCOTUS docket metadata.
        """
        if self.tree is None:
            return {}

        docket_number = None
        docket_title_text = self.tree.xpath(
            "(//table//tr/td[1][starts-with(normalize-space(.), 'No.')])[1]"
        )
        if docket_title_text:
            m = re.search(
                self.DOCKET_NUMBER_RE, docket_title_text[0].text_content()
            )
            docket_number = m.group(1).strip() if m else None

        # Case name
        title_tr = self._htm_row_for_label("Title:")
        title_td = (
            next(iter(title_tr.xpath("./td[2]")), None)
            if title_tr is not None
            else None
        )
        case_name = (
            harmonize(clean_string(self._join_nodes_content([title_td])))
            if title_td is not None
            else ""
        )

        # Dates and lower court info
        date_filed = self.normalize_date(self._htm_value_by_label("Docketed:"))
        lower_court = self._htm_value_by_label("Lower Ct:")

        # Lower court data
        lower_court_raw = self._htm_value_by_label(
            "Case Nos.:", allow_indent=True
        )
        lower_court_raw = (lower_court_raw or "").strip()
        lower_court_cleaned = None
        if lower_court_raw:
            lower_court_cleaned = self.clean_lower_court_cases(
                lower_court_raw.strip("()")
            )
        lower_court_decision_date = self.normalize_date(
            self._htm_value_by_label("Decision Date:", allow_indent=True)
        )
        lower_court_rehearing_denied_date = self.normalize_date(
            self._htm_value_by_label("Rehearing Denied:", allow_indent=True)
        )

        linked_with_label = "string(//tr[td[1][starts-with(normalize-space(.), 'Linked with')]])"
        linked_with_val = self.tree.xpath(linked_with_label).replace(
            "Linked with", ""
        )
        links = (linked_with_val or "").strip()
        return {
            "docket_number": docket_number,
            "capital_case": None,
            "date_filed": date_filed,
            "case_name": case_name or None,
            "links": links,
            "lower_court": lower_court or None,
            "lower_court_case_numbers_raw": lower_court_raw or None,
            "lower_court_case_numbers": lower_court_cleaned,
            "lower_court_decision_date": lower_court_decision_date,
            "lower_court_rehearing_denied_date": lower_court_rehearing_denied_date,
            "questions_presented": None,  # HTM version doesn’t have this link.
            "discretionary_court_decision": None,
        }

    def _build_docket_entry(
        self,
        date_str: str,
        description_td: Optional[HtmlElement],
        attachment_path: str,
    ) -> dict[str, Any]:
        """Build a normalized docket entry dict.

        :param date_str: Raw date string from the date cell.
        :param description_td: The <td> element containing the description.
        :param attachment_path: the path to select attachment anchors inside the
        description cell.
        :return: A dict containing the normalized docket entry.
        """

        description_html = ""
        if description_td is not None:
            # Select content up to first <br> and excluding .documentlinks;
            # fallback to whole cell.
            fragment = (
                self._parse_description_html(description_td).strip()
                or html.tostring(description_td, encoding="unicode").strip()
            )
            # Skip empty tds like <td ...></td>
            if not fragment or re.fullmatch(
                r"<td\b[^>]*>\s*</td>", fragment, flags=re.I
            ):
                description_html = ""
            else:
                description_html = fragment

            # If attachments with links are found in HTM dockets. Log an error
            # This is an opportunity to add support for it.
            if description_td is not None and description_td.xpath(
                attachment_path
            ):
                logger.error("SCOTUS HTM docket entry contains attachments.")

        description = ""
        raw = (description_html or "").strip()
        if raw:
            node = strip_bad_html_tags_insecure(raw)
            description = self._clean_whitespace(node.text_content())

        return {
            "date_filed": self.normalize_date(date_str),
            "description": description,
            "description_html": description_html,
            "attachments": [],
        }

    @property
    def docket_entries(self) -> list[dict[str, Any]]:
        """Return docket entries from the HTM 'Proceedings and Orders' table.

        :return: List of dicts with date_filed, description, description_html, attachments.
        """
        if self.tree is None:
            return []

        tables = self.tree.xpath(
            "//table[.//tr[td[1][contains(normalize-space(.), '~~~Date~~~')] "
            "and td[2][contains(normalize-space(.), 'Proceedings')]]]"
        )
        if not tables:
            return []

        header_tr = next(
            iter(
                tables[0].xpath(
                    ".//tr[td[1][contains(normalize-space(.), '~~~Date~~~')] "
                    "and td[2][contains(normalize-space(.), 'Proceedings')]]"
                )
            ),
            None,
        )
        if header_tr is None:
            return []

        entries = []

        # Only iterate rows after the header within the same table
        for tr in header_tr.itersiblings(tag="tr"):
            tds = tr.xpath("./td")
            if not tds:
                continue
            date_str = self._clean_whitespace(tds[0].text_content())
            desc_td = tds[1]

            # Build the entry
            entries.append(
                self._build_docket_entry(
                    date_str=date_str,
                    description_td=desc_td,
                    attachment_path=".//a[@href]",
                )
            )

        return entries

    def _build_htm_attorney(
        self,
        current_attorney: Optional[dict[str, Any]],
    ) -> Optional[dict[str, Any]]:
        """Complete an attorney dict from the HTM Contacts table.

        :param current_attorney: The in-progress attorney dict with initial values.
        :return: The current_attorney dict.
        """
        if current_attorney is None:
            return None

        lines = current_attorney.pop("_raw_lines", [])
        email = current_attorney.pop("_email", None)

        # Filter lines: remove lines with IDs like "#1098260" and inline email if present.
        filtered_lines = []
        for ln in lines:
            if email is None:
                m = self.EMAIL_RE.search(ln)
                if m:
                    email = m.group(0)
                    continue
            if self.ID_RE.search(ln):
                continue
            if ln:
                filtered_lines.append(ln)

        title, start_add_idx = self._parse_address_title(filtered_lines)
        partial_address = self._parse_address_lines(
            filtered_lines, start_add_idx
        )

        addr_lines = partial_address.address_lines
        current_attorney.update(
            {
                "title": title,
                "address": ", ".join(addr_lines) if addr_lines else None,
                "city": partial_address.city,
                "state": partial_address.state,
                "zip": partial_address.zip_code,
                "email": email,
            }
        )
        return current_attorney

    @property
    def parties(self) -> list[dict[str, Any]]:
        """Return parties grouped by type for HTM pages.

        :return: List of dicts containing parties data.
        """
        if self.tree is None:
            return []

        # Locate the contacts table.
        contacts_tbl = next(
            iter(self.tree.xpath(self.CONTACTS_TABLE_XPATH)), None
        )
        if contacts_tbl is None:
            return []

        header_tr = next(iter(contacts_tbl.xpath(self.HEADER_ROW_XPATH)), None)
        if header_tr is None:
            return []

        parties_by_key = defaultdict(list)
        current_type = None
        current_attorney = None

        for tr in header_tr.itersiblings(tag="tr"):
            # Match table header e.g: "Attorneys for Petitioner:"
            header_text = self._clean_whitespace(
                "".join(tr.xpath("./td[1]//b/text()"))
            )
            if header_text:
                # Inside Petitioner, Respondent or Other parties.
                current_type = self._map_contacts_header_to_type(header_text)
                continue

            # Party name row. Finish previous attorney.
            party_name_re = r"^Party name:\s*"
            party_line = self._clean_whitespace(tr.text_content())
            if re.search(party_name_re, party_line, flags=re.I):
                current_party_name = re.sub(
                    party_name_re, "", party_line, flags=re.I
                ).strip()
                if current_attorney:
                    current_attorney = self._build_htm_attorney(
                        current_attorney
                    )
                    type_key = (current_type or "Other", current_party_name)
                    parties_by_key[type_key].append(current_attorney)
                continue

            tds = tr.xpath("./td")
            if not tds:
                continue

            # Extract normalized columns.
            name_col = (
                self._clean_whitespace(tds[0].text_content())
                if len(tds) >= 1
                else ""
            )
            address_col = (
                self._clean_whitespace(tds[1].text_content())
                if len(tds) >= 2
                else ""
            )
            phone_col = (
                self._clean_whitespace(tds[2].text_content())
                if len(tds) >= 3
                else None
            )

            # Email-only row appears in the address column.
            if (
                not name_col
                and address_col
                and self.EMAIL_RE.search(address_col)
            ):
                if current_attorney:
                    current_attorney["_email"] = self.EMAIL_RE.search(
                        address_col
                    ).group(0)
                continue

            # "Counsel of Record" marker row.
            if re.search(r"\bcounsel\s+of\s+record\b", name_col, re.I):
                if current_attorney:
                    current_attorney["is_counsel_of_record"] = True
                    if address_col:
                        current_attorney.setdefault("_raw_lines", []).append(
                            address_col
                        )
                continue

            # Beginning of a new attorney row.
            if name_col:
                current_attorney = {
                    "name": name_col or None,
                    "is_counsel_of_record": False,
                    "phone": phone_col or None,
                    "_raw_lines": [],
                    "_email": None,
                }
                if address_col:
                    current_attorney["_raw_lines"].append(address_col)
                continue

            # Address parsing continuation for the current attorney.
            if not name_col and address_col and current_attorney:
                current_attorney["_raw_lines"].append(address_col)
                continue

        return [
            {"type": party_type, "name": name, "attorneys": attorneys}
            for (party_type, name), attorneys in parties_by_key.items()
        ]

    def _htm_row_for_label(
        self, label: str, allow_indent: bool = False
    ) -> Optional[HtmlElement]:
        """Return the <tr> whose first <td> equals the given label.

        :param label: The exact label text, e.g. "Docketed:" or "Case Nos.:".
        :param allow_indent: If True, normalize non-breaking spaces before matching.
        :return: The <tr> element or None.
        """
        if self.tree is None:
            return None

        if allow_indent:
            # Allow \u00a0 for the first td cell text
            target_label = f"//tr[td][normalize-space(translate(string(td[1]), '\u00a0',' '))='{label}']"
        else:
            target_label = (
                f"//tr[td][normalize-space(string(td[1]))='{label}']"
            )

        return next(iter(self.tree.xpath(target_label)), None)

    def _htm_value_by_label(
        self, label: str, allow_indent: bool = False
    ) -> Optional[str]:
        """Get the value text in td[2] for a labeled row.

        :param label: The label text in td[1].
        :param allow_indent: True to allow labels with leading &nbsp;.
        :return: Cleaned value string or None.
        """
        tr = self._htm_row_for_label(label, allow_indent=allow_indent)
        if tr is None:
            return None
        val_td = next(iter(tr.xpath("./td[2]")), None)
        return (
            self._clean_whitespace(val_td.text_content())
            if val_td is not None
            else None
        )


def _main():
    """Parse a local HTML file and pretty-print normalized data.

    :return: None
    """
    import pprint
    import sys

    if len(sys.argv) != 2:
        print(
            "Usage: python -m juriscraper.scotus_docket_report_htm <filepath>"
        )
        sys.exit(1)

    parser = SCOTUSDocketReportHTM()
    with open(sys.argv[1], encoding="utf-8") as f:
        parser._parse_text(f.read())
    pprint.pprint(parser.data, indent=2)


if __name__ == "__main__":
    _main()



from datetime import datetime

from juriscraper.opinions.united_states.state import alaska
from juriscraper.OpinionSite import OpinionSite


class Site(alaska.Site):
    # In the Matter of the Suspension of Marcus B Paine v Opinion Number 35
    first_opinion_date = datetime(2000, 6, 30).date()
    is_coa = False
    search_parameter = "PublishedOrders"
    document_type = "PUBORDER"
    extract_from_text = OpinionSite.extract_from_text

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



"""Scraper for the Maryland Attorney General
CourtID: mdag
Court Short Name: Maryland Attorney General
"""

import re
from datetime import date

from dateutil import parser

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://oag.maryland.gov/resources-info/_api/web/lists/getbytitle('Search-Opinions')/items?$top=100&$select=Title,Year,Summary,Attachment,Created_x0020_Date&$orderby=Year desc"
        self.year = date.today().year

        self.method = "GET"
        self.request["headers"] = {
            "Accept": "application/json;odata=verbose",
        }
        self.needs_special_headers = True
        self.status = "Published"

    def _process_html(self):
        self.json = self.html
        for row in self.json["d"]["results"]:
            docket_number = row["Title"]
            title = "Maryland Attorney General Opinion " + docket_number
            year = row["Year"]
            attachment = row["Attachment"]
            approximate_date = f"{year}-01-01"
            summary = row["Summary"]

            citation = ""
            citation_pattern = r"(\d+)\s*OAG\s*(\d+)"
            if match := re.search(citation_pattern, docket_number):
                citation = (
                    f"{match.group(1)} Op. Atty Gen. Md. {match.group(2)}"
                )

            self.cases.append(
                {
                    "docket": docket_number,
                    "name": title,
                    "summary": summary,
                    "url": f"https://oag.maryland.gov/resources-info/Documents/pdfs/Opinions/{year}/{attachment}.pdf",
                    "date": approximate_date,
                    "date_filed_is_approximate": True,
                    "citation": citation,
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract the exact date(s) from the scraped text."""
        pattern = r"([A-Z][a-z]+ \d{1,2}, \d{4})"

        if matches := re.findall(pattern, scraped_text):
            date_string = parser.parse(matches[0]).strftime("%Y-%m-%d")
            return {
                "OpinionCluster": {
                    "date_filed": date_string,
                    "date_filed_is_approximate": False,
                }
            }
        return {}



"""
Scraper for Indiana Court of Appeals Reclassified Opinions
CourtID: indctapp
Court Short Name: Ind. Ct. App.
Auth: Luis Manzur
History:
    2025-07-29: Created by Luis Manzur
"""

from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.status = "Published"
        self.url = "https://www.in.gov/courts/appeals/reclassified/"
        self.should_have_results = True
        self.expected_content_types = ["application/pdf"]

    def _process_html(self):
        """Process the HTML to extract case details.

        :return None
        """

        rows = self.html.xpath("//table/tbody/tr")
        for row in rows:
            cells = row.xpath(".//td")
            link = cells[2].xpath(".//a")
            if not link:
                continue

            name_parts = link[0].xpath(".//p")
            if name_parts:
                name = titlecase(" ".join(p.text.strip() for p in name_parts))
            else:
                name = titlecase(link[0].text.strip())
            url = link[0].get("href")

            lower_court_parts = cells[3].xpath(".//p")
            if lower_court_parts:
                lower_court = ", ".join(
                    p.text.strip() for p in lower_court_parts
                )
            else:
                lower_court = ", ".join(
                    part.replace(",", ", ").strip()
                    for part in cells[3].xpath(".//text()[normalize-space()]")
                    if part.strip()
                ).strip(", ")

            docket_parts = cells[4].xpath(".//p")
            docket = (
                docket_parts[0].text.strip()
                if docket_parts
                else cells[4].text.strip()
            )

            self.cases.append(
                {
                    "date": cells[1].text.strip(),
                    "other_date": f"Date opinion was reclassified as published: {cells[0].text.strip()}",
                    "name": name,
                    "url": url,
                    "lower_court_number": lower_court,
                    "docket": docket,
                }
            )



import logging
from typing import Optional

from juriscraper.lib.log_tools import make_default_logger

logger = make_default_logger()


class JuriscraperException(Exception):
    """
    Base class for Juriscraper custom exceptions
    """

    logger = logger


class SkipRowError(JuriscraperException):
    """
    Raise when a row or record has to be skipped
    For example, when collecting opinions and finding an order
    """

    def __init__(self, message: str):
        self.logger.debug(message)


class SlownessException(Exception):
    """Raised when things are too slow."""

    def __init__(self, message):
        Exception.__init__(self, message)


class ParsingException(Exception):
    """Raised when parsing fails."""

    def __init__(self, message):
        Exception.__init__(self, message)


class InsanityException(Exception):
    """Raised when data validation fails."""

    def __init__(self, message):
        Exception.__init__(self, message)


class PacerLoginException(Exception):
    """Raised when the system cannot authenticate with PACER"""

    def __init__(self, message):
        Exception.__init__(self, message)


class InvalidDocumentError(Exception):
    """Raised when the document got from `download_url` is invalid

    May be an error page that is undetected by `response.raise_for_status`
    or our `expected_content_type` controls. Proper place to raise this
    would be on `Site.cleanup_content`
    """


class AutoLoggingException(Exception):
    """Exception with defaults for logging, to be subclassed

    We log expected exceptions to better understand what went wrong
    Logger calls with level `logging.ERROR` are sent to Sentry, and
    it's useful to send a `fingerprint` to force a specific grouping by court

    Other `logger` calls are just printed on the console when using a
    VerboseCommand with proper verbosity levels
    """

    logging_level = logging.DEBUG
    message = ""
    logger = logger

    def __init__(
        self,
        message: str = "",
        logger: Optional[logging.Logger] = None,
        logging_level: Optional[int] = None,
        fingerprint: Optional[list[str]] = None,
    ):
        if not message:
            message = self.message
        if not logger:
            logger = self.logger
        if not logging_level:
            logging_level = self.logging_level

        log_kwargs = {}
        if fingerprint:
            log_kwargs["extra"] = {"fingerprint": fingerprint}

        logger.log(logging_level, message, **log_kwargs)
        super().__init__(message)


class BadContentError(AutoLoggingException):
    """Parent class for errors raised when downloading binary content"""


class UnexpectedContentTypeError(BadContentError):
    """Occurs when the content received from the server has
    a different content type than the ones listed on
    site.expected_content_types
    """

    logging_level = logging.ERROR


class NoDownloadUrlError(BadContentError):
    """Occurs when a DeferredList fetcher fails."""

    logging_level = logging.ERROR


class EmptyFileError(BadContentError):
    """Occurs when the content of the response has lenght 0"""

    logging_level = logging.ERROR


class MergingError(AutoLoggingException):
    """Raised when metadata merging finds different values"""

    logging_level = logging.ERROR



from datetime import datetime

from juriscraper.opinions.united_states.state import alaska


class Site(alaska.Site):
    # Charles E. Burnett v Anchorage Refuse, Inc., et al. Opinion Number 846
    first_opinion_date = datetime(1997, 3, 5).date()
    is_coa = False
    search_parameter = "MOJOpinions"
    document_type = "MO&J"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status = "Unpublished"
        self.is_citation_visible = False
